/***********************************************************************
** Copyright (C) 2003  ACX100 Open Source Project
**
** The contents of this file are subject to the Mozilla Public
** License Version 1.1 (the "License"); you may not use this file
** except in compliance with the License. You may obtain a copy of
** the License at http://www.mozilla.org/MPL/
**
** Software distributed under the License is distributed on an "AS
** IS" basis, WITHOUT WARRANTY OF ANY KIND, either express or
** implied. See the License for the specific language governing
** rights and limitations under the License.
**
** Alternatively, the contents of this file may be used under the
** terms of the GNU Public License version 2 (the "GPL"), in which
** case the provisions of the GPL are applicable instead of the
** above.  If you wish to allow the use of your version of this file
** only under the terms of the GPL and not to allow others to use
** your version of this file under the MPL, indicate your decision
** by deleting the provisions above and replace them with the notice
** and other provisions required by the GPL.  If you do not delete
** the provisions above, a recipient may use your version of this
** file under either the MPL or the GPL.
** ---------------------------------------------------------------------
** Inquiries regarding the ACX100 Open Source Project can be
** made directly to:
**
** acx100-users@lists.sf.net
** http://acx100.sf.net
** ---------------------------------------------------------------------
*/

#include <linux/config.h>
#include <linux/version.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/types.h>
#include <linux/timer.h>
#include <linux/slab.h>
#include <linux/proc_fs.h>
#include <linux/if_arp.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/rtnetlink.h>
#include <linux/wireless.h>
#if WIRELESS_EXT >= 13
#include <net/iw_handler.h>
#endif
#include <linux/pm.h>

#include "acx.h"


/***********************************************************************
*/

/* Now that the pci_alloc_consistent() problem has been resolved,
 * feel free to modify buffer count for ACX100 to 32, too.
 * But it's not required since the card isn't too fast anyway */
#define RXBUFCNT_ACX100 16
#define TXBUFCNT_ACX100 16
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 5, 53)
/* dma_alloc_coherent() uses GFP_KERNEL, much less problematic than
 * the pci_alloc_consistent() used below using GFP_ATOMIC (quite often causes
 * a larger alloc to fail), so use less buffers there to be more successful */
#define RXBUFCNT_ACX111 32
#define TXBUFCNT_ACX111 32
#else
#define RXBUFCNT_ACX111 16
#define TXBUFCNT_ACX111 16
#endif

/* Probably a number of acx's itermediate buffers for USB transfers,
** not to be confused with number of descriptors in tx/rx rings
** (which are not directly accessible to host in USB devices) */
#define USB_RXBUFCNT 10
#define USB_TXBUFCNT 10


/*----------------------------------------------------------------
* acx100_s_init_memory_pools
*----------------------------------------------------------------*/
void BUG_acx100_ie_memconfigoption_must_be_24_bytes_in_length(void);

static int
acx100_s_init_memory_pools(wlandevice_t *priv, const acx_ie_memmap_t *mmt)
{
	acx100_ie_memblocksize_t MemoryBlockSize;
	acx100_ie_memconfigoption_t MemoryConfigOption;
	int TotalMemoryBlocks;
	int RxBlockNum;
	int TotalRxBlockSize;
	int TxBlockNum;
	int TotalTxBlockSize;

	if (sizeof(acx100_ie_memconfigoption_t) != 24)
		BUG_acx100_ie_memconfigoption_must_be_24_bytes_in_length();

	FN_ENTER;

	/* Let's see if we can follow this:
	   first we select our memory block size (which I think is
	   completely arbitrary) */
	MemoryBlockSize.size = cpu_to_le16(priv->memblocksize);

	/* Then we alert the card to our decision of block size */
	if (OK != acx_s_configure(priv, &MemoryBlockSize, ACX100_IE_BLOCK_SIZE)) {
		goto bad;
	}

	/* We figure out how many total blocks we can create, using
	   the block size we chose, and the beginning and ending
	   memory pointers, i.e.: end-start/size */
	TotalMemoryBlocks = (le32_to_cpu(mmt->PoolEnd) - le32_to_cpu(mmt->PoolStart)) / priv->memblocksize;

	acxlog(L_DEBUG, "TotalMemoryBlocks=%u (%u bytes)\n",
		TotalMemoryBlocks, TotalMemoryBlocks*priv->memblocksize);

	/* MemoryConfigOption.DMA_config bitmask:
			// access to ACX memory is to be done:
	0x00080000	//  using PCI conf space?!
	0x00040000	//  using IO instructions?
	0x00000000	//  using memory access instructions
	0x00020000	// use local memory block linked list (else what?)
	0x00010000	// use host indirect descriptors (else host must access ACX memory?)
	*/
	if (IS_PCI(priv)) {
		MemoryConfigOption.DMA_config = cpu_to_le32(0x30000);
		/* Declare start of the Rx host pool */
//// this one comes from PCI hostdesc init:
		MemoryConfigOption.pRxHostDesc = ptr2acx(priv->RxHostDescPoolStart);
		acxlog(L_DEBUG, "pRxHostDesc 0x%08X, RxHostDescPoolStart 0x%p\n",
				acx2cpu(MemoryConfigOption.pRxHostDesc),
				priv->RxHostDescPoolStart);
	} else {
		MemoryConfigOption.DMA_config = cpu_to_le32(0x20000);
	}

	/* 50% of the allotment of memory blocks go to tx descriptors */
	TxBlockNum = TotalMemoryBlocks / 2;
	MemoryConfigOption.TxBlockNum = cpu_to_le16(TxBlockNum);

	/* and 50% go to the rx descriptors */
	RxBlockNum = TotalMemoryBlocks - TxBlockNum;
	MemoryConfigOption.RxBlockNum = cpu_to_le16(RxBlockNum);

	/* size of the tx and rx descriptor queues */
	TotalTxBlockSize = TxBlockNum * priv->memblocksize;
	TotalRxBlockSize = RxBlockNum * priv->memblocksize;
	acxlog(L_DEBUG, "TxBlockNum %u RxBlockNum %u TotalTxBlockSize %u "
		"TotalTxBlockSize %u\n", TxBlockNum, RxBlockNum,
		TotalTxBlockSize, TotalRxBlockSize);


	/* align the tx descriptor queue to an alignment of 0x20 (32 bytes) */
	MemoryConfigOption.rx_mem =
		cpu_to_le32((le32_to_cpu(mmt->PoolStart) + 0x1f) & ~0x1f);

	/* align the rx descriptor queue to units of 0x20
	 * and offset it by the tx descriptor queue */
	MemoryConfigOption.tx_mem =
		cpu_to_le32((le32_to_cpu(mmt->PoolStart) + TotalRxBlockSize + 0x1f) & ~0x1f);
	acxlog(L_DEBUG, "rx_mem %08X rx_mem %08X\n",
		MemoryConfigOption.tx_mem, MemoryConfigOption.rx_mem);

	/* alert the device to our decision */
	if (OK != acx_s_configure(priv, &MemoryConfigOption, ACX1xx_IE_MEMORY_CONFIG_OPTIONS)) {
		goto bad;
	}

	/* and tell the device to kick it into gear */
	if (OK != acx_s_issue_cmd(priv, ACX100_CMD_INIT_MEMORY, NULL, 0)) {
		goto bad;
	}
	FN_EXIT1(OK);
	return OK;
bad:
	FN_EXIT1(NOT_OK);
	return NOT_OK;
}


/*----------------------------------------------------------------
* acx100_s_create_dma_regions
*
* Note that this fn messes up heavily with hardware, but we cannot
* lock it (we need to sleep). Not a problem since IRQs can't happen
*----------------------------------------------------------------*/
void BUG_acx100_ie_queueconfig_t_must_be_0x20_bytes_in_length(void);

int
acx100_s_create_dma_regions(wlandevice_t *priv)
{
	acx100_ie_queueconfig_t queueconf;
	acx_ie_memmap_t memmap;
	int res = NOT_OK;
	u32 tx_queue_start, rx_queue_start;

	FN_ENTER;

	/* read out the acx100 physical start address for the queues */
	if (OK != acx_s_interrogate(priv, &memmap, ACX1xx_IE_MEMORY_MAP)) {
		goto fail;
	}

	/* # of items in Rx and Tx queues */
	priv->TxQueueCnt = TXBUFCNT_ACX100;
	priv->RxQueueCnt = RXBUFCNT_ACX100;

	tx_queue_start = le32_to_cpu(memmap.QueueStart);
	rx_queue_start = tx_queue_start + TXBUFCNT_ACX100 * sizeof(txdesc_t);

	acxlog(L_DEBUG, "initializing Queue Indicator\n");

	memset(&queueconf, 0, sizeof(queueconf));

	/* Not needed for PCI, so we can avoid setting them altogether */
	if (IS_USB(priv)) {
		queueconf.NumTxDesc = USB_TXBUFCNT;
		queueconf.NumRxDesc = USB_RXBUFCNT;
	}

	/* calculate size of queues */
	queueconf.AreaSize = cpu_to_le32(
			TXBUFCNT_ACX100 * sizeof(txdesc_t) +
			RXBUFCNT_ACX100 * sizeof(rxdesc_t) + 8
			);
	queueconf.NumTxQueues = 1;  /* number of tx queues */
	/* sets the beginning of the tx descriptor queue */
	queueconf.TxQueueStart = memmap.QueueStart;
	/* done by memset: queueconf.TxQueuePri = 0; */
	queueconf.RxQueueStart = cpu_to_le32(rx_queue_start);
	queueconf.QueueOptions = 1;		/* auto reset descriptor */
	/* sets the end of the rx descriptor queue */
	queueconf.QueueEnd = cpu_to_le32(
			rx_queue_start + RXBUFCNT_ACX100 * sizeof(rxdesc_t)
			);
	/* sets the beginning of the next queue */
	queueconf.HostQueueEnd = cpu_to_le32(le32_to_cpu(queueconf.QueueEnd) + 8);
	if (sizeof(queueconf) != 0x20)
		BUG_acx100_ie_queueconfig_t_must_be_0x20_bytes_in_length();
	if (OK != acx_s_configure(priv, &queueconf, ACX1xx_IE_QUEUE_CONFIG)) {
		goto fail;
	}

	if (IS_PCI(priv)) {
	/* sets the beginning of the rx descriptor queue, after the tx descrs */
		if (OK != acx_s_create_hostdesc_queues(priv))
			goto fail;
		acx_create_desc_queues(priv, tx_queue_start, rx_queue_start);
	}

	if (OK != acx_s_interrogate(priv, &memmap, ACX1xx_IE_MEMORY_MAP)) {
		goto fail;
	}

/* [20050901] seems to be bogus. remove if no one complains */
#if 0 /* #ifdef ACX_USB */
	if (OK != acx_s_configure(priv, &memmap, ACX1xx_IE_MEMORY_MAP)) {
		goto fail;
	}
#endif

	memmap.PoolStart = cpu_to_le32(
			(le32_to_cpu(memmap.QueueEnd) + 4 + 0x1f) & ~0x1f
			);

	if (OK != acx_s_configure(priv, &memmap, ACX1xx_IE_MEMORY_MAP)) {
		goto fail;
	}

	if (OK != acx100_s_init_memory_pools(priv, &memmap)) {
		goto fail;
	}

	res = OK;
	goto end;

fail:
	acx_s_msleep(1000); /* ? */
	if (IS_PCI(priv))
		acx_free_desc_queues(priv);
end:
	FN_EXIT1(res);
	return res;
}


/*----------------------------------------------------------------
* acx111_s_create_dma_regions
*
* Note that this fn messes up heavily with hardware, but we cannot
* lock it (we need to sleep). Not a problem since IRQs can't happen
*----------------------------------------------------------------*/
#define ACX111_PERCENT(percent) ((percent)/5)

int
acx111_s_create_dma_regions(wlandevice_t *priv)
{
	struct acx111_ie_memoryconfig memconf;
	struct acx111_ie_queueconfig queueconf;
	u32 tx_queue_start, rx_queue_start;

	FN_ENTER;

	/* Calculate memory positions and queue sizes */

	priv->TxQueueCnt = TXBUFCNT_ACX111;
	priv->RxQueueCnt = RXBUFCNT_ACX111;

	/* Set up our host descriptor pool + data pool */
	if (IS_PCI(priv)) {
		if (OK != acx_s_create_hostdesc_queues(priv))
			goto fail;
	}

	memset(&memconf, 0, sizeof(memconf));
	/* the number of STAs (STA contexts) to support
	** NB: was set to 1 and everything seemed to work nevertheless... */
	memconf.no_of_stations = cpu_to_le16(VEC_SIZE(priv->sta_list));
	/* specify the memory block size. Default is 256 */
	memconf.memory_block_size = cpu_to_le16(priv->memblocksize);
	/* let's use 50%/50% for tx/rx (specify percentage, units of 5%) */
	memconf.tx_rx_memory_block_allocation = ACX111_PERCENT(50);
	/* set the count of our queues
	** NB: struct acx111_ie_memoryconfig shall be modified
	** if we ever will switch to more than one rx and/or tx queue */
	memconf.count_rx_queues = 1;
	memconf.count_tx_queues = 1;
	/* 0 == Busmaster Indirect Memory Organization, which is what we want
	 * (using linked host descs with their allocated mem).
	 * 2 == Generic Bus Slave */
	/* done by memset: memconf.options = 0; */
	/* let's use 25% for fragmentations and 75% for frame transfers
	 * (specified in units of 5%) */
	memconf.fragmentation = ACX111_PERCENT(75);
	/* Rx descriptor queue config */
	memconf.rx_queue1_count_descs = RXBUFCNT_ACX111;
	memconf.rx_queue1_type = 7; /* must be set to 7 */
	/* done by memset: memconf.rx_queue1_prio = 0; low prio */
//// this one comes from PCI hostdesc init. BUG! we need aligned one!
///	memconf.rx_queue1_host_rx_start = cpu_to_le32(priv->RxHostDescQPoolPhyAddr);
	memconf.rx_queue1_host_rx_start = cpu_to_le32(priv->RxHostDescPoolStart);
	/* Tx descriptor queue config */
	memconf.tx_queue1_count_descs = TXBUFCNT_ACX111;
	/* done by memset: memconf.tx_queue1_attributes = 0; lowest priority */

	/* NB1: this looks wrong: (memconf,ACX1xx_IE_QUEUE_CONFIG),
	** (queueconf,ACX1xx_IE_MEMORY_CONFIG_OPTIONS) look swapped, eh?
	** But it is actually correct wrt IE numbers.
	** NB2: sizeof(memconf) == 28 == 0x1c but configure(ACX1xx_IE_QUEUE_CONFIG)
	** writes 0x20 bytes (because same IE for acx100 uses struct acx100_ie_queueconfig
	** which is 4 bytes larger. what a mess. TODO: clean it up) */
	if (OK != acx_s_configure(priv, &memconf, ACX1xx_IE_QUEUE_CONFIG)) {
		goto fail;
	}

	acx_s_interrogate(priv, &queueconf, ACX1xx_IE_MEMORY_CONFIG_OPTIONS);

	tx_queue_start = le32_to_cpu(queueconf.tx1_queue_address);
	rx_queue_start = le32_to_cpu(queueconf.rx1_queue_address);

	acxlog(L_INIT, "dump queue head (from card):\n"
		       "len: %u\n"
		       "tx_memory_block_address: %X\n"
		       "rx_memory_block_address: %X\n"
		       "tx1_queue address: %X\n"
		       "rx1_queue address: %X\n",
		       le16_to_cpu(queueconf.len),
		       le32_to_cpu(queueconf.tx_memory_block_address),
		       le32_to_cpu(queueconf.rx_memory_block_address),
		       tx_queue_start,
		       rx_queue_start);
	if (IS_PCI(priv))
		acx_create_desc_queues(priv, tx_queue_start, rx_queue_start);

	FN_EXIT1(OK);
	return OK;
fail:
	if (IS_PCI(priv))
		acx_free_desc_queues(priv);

	FN_EXIT1(NOT_OK);
	return NOT_OK;
}


#ifdef CONFIG_PROC_FS
/***********************************************************************
** /proc files
*/
/***********************************************************************
** acx_l_proc_output
** Generate content for our /proc entry
**
** Arguments:
**	buf is a pointer to write output to
**	priv is the usual pointer to our private struct wlandevice
** Returns:
**	number of bytes actually written to buf
** Side effects:
**	none
*/
static int
acx_l_proc_output(char *buf, wlandevice_t *priv)
{
	char *p = buf;
	int i;

	FN_ENTER;

	p += sprintf(p,
		"acx driver version:\t\t" WLAN_RELEASE "\n"
		"Wireless extension version:\t" STRING(WIRELESS_EXT) "\n"
		"chip name:\t\t\t%s (0x%08X)\n"
		"radio type:\t\t\t0x%02X\n"
		/* TODO: add radio type string from acx_display_hardware_details */
		"form factor:\t\t\t0x%02X\n"
		/* TODO: add form factor string from acx_display_hardware_details */
		"EEPROM version:\t\t\t0x%02X\n"
		"firmware version:\t\t%s (0x%08X)\n",
		priv->chip_name, priv->firmware_id,
		priv->radio_type,
		priv->form_factor,
		priv->eeprom_version,
		priv->firmware_version, priv->firmware_numver);

	for (i = 0; i < VEC_SIZE(priv->sta_list); i++) {
		struct client *bss = &priv->sta_list[i];
		if (!bss->used) continue;
		p += sprintf(p, "BSS %u BSSID "MACSTR" ESSID %s channel %u "
			"Cap 0x%X SIR %u SNR %u\n",
			i, MAC(bss->bssid), (char*)bss->essid, bss->channel,
			bss->cap_info, bss->sir, bss->snr);
	}
	p += sprintf(p, "status:\t\t\t%u (%s)\n",
			priv->status, acx_get_status_name(priv->status));
	/* TODO: add more interesting stuff (essid, ...) here */

	FN_EXIT1(p - buf);
	return p - buf;
}


/***********************************************************************
*/
static int
acx_s_proc_diag_output(char *buf, wlandevice_t *priv)
{
	char *p = buf;
	fw_stats_t *fw_stats;
	unsigned long flags;

	FN_ENTER;

	fw_stats = kmalloc(sizeof(fw_stats_t), GFP_KERNEL);
	if (!fw_stats) {
		FN_EXIT1(0);
		return 0;
	}
	memset(fw_stats, 0, sizeof(fw_stats_t));

	acx_lock(priv, flags);

	if (IS_PCI(priv))
		p = acxpci_s_proc_diag_output(p, priv);

	p += sprintf(p,
		"\n"
		"** network status **\n"
		"dev_state_mask 0x%04X\n"
		"status %u (%s), "
		"mode %u, channel %u, "
		"reg_dom_id 0x%02X, reg_dom_chanmask 0x%04X, ",
		priv->dev_state_mask,
		priv->status, acx_get_status_name(priv->status),
		priv->mode, priv->channel,
		priv->reg_dom_id, priv->reg_dom_chanmask
		);
	p += sprintf(p,
		"ESSID \"%s\", essid_active %d, essid_len %d, "
		"essid_for_assoc \"%s\", nick \"%s\"\n"
		"WEP ena %d, restricted %d, idx %d\n",
		priv->essid, priv->essid_active, (int)priv->essid_len,
		priv->essid_for_assoc, priv->nick,
		priv->wep_enabled, priv->wep_restricted,
		priv->wep_current_index);
	p += sprintf(p, "dev_addr  "MACSTR"\n", MAC(priv->dev_addr));
	p += sprintf(p, "bssid     "MACSTR"\n", MAC(priv->bssid));
	p += sprintf(p, "ap_filter "MACSTR"\n", MAC(priv->ap));

	p += sprintf(p,
		"\n"
		"** PHY status **\n"
		"tx_disabled %d, tx_level_dbm %d\n" /* "tx_level_val %d, tx_level_auto %d\n" */
		"sensitivity %d, antenna 0x%02X, ed_threshold %d, cca %d, preamble_mode %d\n"
		"rts_threshold %d, short_retry %d, long_retry %d, msdu_lifetime %d, listen_interval %d, beacon_interval %d\n",
		priv->tx_disabled, priv->tx_level_dbm, /* priv->tx_level_val, priv->tx_level_auto, */
		priv->sensitivity, priv->antenna, priv->ed_threshold, priv->cca, priv->preamble_mode,
		priv->rts_threshold, priv->short_retry, priv->long_retry, priv->msdu_lifetime, priv->listen_interval, priv->beacon_interval);

	acx_unlock(priv, flags);

	if (OK != acx_s_interrogate(priv, fw_stats, ACX1xx_IE_FIRMWARE_STATISTICS))
		p += sprintf(p,
			"\n"
			"** Firmware **\n"
			"QUERY FAILED!!\n");
	else {
		p += sprintf(p,
			"\n"
			"** Firmware **\n"
			"version \"%s\"\n"
			"tx_desc_overfl %u, rx_OutOfMem %u, rx_hdr_overfl %u, rx_hdr_use_next %u\n"
			"rx_dropped_frame %u, rx_frame_ptr_err %u, rx_xfr_hint_trig %u, rx_dma_req %u\n"
			"rx_dma_err %u, tx_dma_req %u, tx_dma_err %u, cmd_cplt %u, fiq %u\n"
			"rx_hdrs %u, rx_cmplt %u, rx_mem_overfl %u, rx_rdys %u, irqs %u\n"
			"acx_trans_procs %u, decrypt_done %u, dma_0_done %u, dma_1_done %u\n",
			priv->firmware_version,
			le32_to_cpu(fw_stats->tx_desc_of),
			le32_to_cpu(fw_stats->rx_oom),
			le32_to_cpu(fw_stats->rx_hdr_of),
			le32_to_cpu(fw_stats->rx_hdr_use_next),
			le32_to_cpu(fw_stats->rx_dropped_frame),
			le32_to_cpu(fw_stats->rx_frame_ptr_err),
			le32_to_cpu(fw_stats->rx_xfr_hint_trig),
			le32_to_cpu(fw_stats->rx_dma_req),
			le32_to_cpu(fw_stats->rx_dma_err),
			le32_to_cpu(fw_stats->tx_dma_req),
			le32_to_cpu(fw_stats->tx_dma_err),
			le32_to_cpu(fw_stats->cmd_cplt),
			le32_to_cpu(fw_stats->fiq),
			le32_to_cpu(fw_stats->rx_hdrs),
			le32_to_cpu(fw_stats->rx_cmplt),
			le32_to_cpu(fw_stats->rx_mem_of),
			le32_to_cpu(fw_stats->rx_rdys),
			le32_to_cpu(fw_stats->irqs),
			le32_to_cpu(fw_stats->acx_trans_procs),
			le32_to_cpu(fw_stats->decrypt_done),
			le32_to_cpu(fw_stats->dma_0_done),
			le32_to_cpu(fw_stats->dma_1_done));
		p += sprintf(p,
			"tx_exch_complet %u, commands %u, acx_rx_procs %u\n"
			"hw_pm_mode_changes %u, host_acks %u, pci_pm %u, acm_wakeups %u\n"
			"wep_key_count %u, wep_default_key_count %u, dot11_def_key_mib %u\n"
			"wep_key_not_found %u, wep_decrypt_fail %u\n",
			le32_to_cpu(fw_stats->tx_exch_complet),
			le32_to_cpu(fw_stats->commands),
			le32_to_cpu(fw_stats->acx_rx_procs),
			le32_to_cpu(fw_stats->hw_pm_mode_changes),
			le32_to_cpu(fw_stats->host_acks),
			le32_to_cpu(fw_stats->pci_pm),
			le32_to_cpu(fw_stats->acm_wakeups),
			le32_to_cpu(fw_stats->wep_key_count),
			le32_to_cpu(fw_stats->wep_default_key_count),
			le32_to_cpu(fw_stats->dot11_def_key_mib),
			le32_to_cpu(fw_stats->wep_key_not_found),
			le32_to_cpu(fw_stats->wep_decrypt_fail));
	}

	kfree(fw_stats);

	FN_EXIT1(p - buf);
	return p - buf;
}


/***********************************************************************
*/
static int
acx_s_proc_phy_output(char *buf, wlandevice_t *priv)
{
	char *p = buf;
	int i;

	FN_ENTER;

	/*
	if (RADIO_RFMD_11 != priv->radio_type) {
		printk("sorry, not yet adapted for radio types "
			"other than RFMD, please verify "
			"PHY size etc. first!\n");
		goto end;
	}
	*/

	/* The PHY area is only 0x80 bytes long; further pages after that
	 * only have some page number registers with altered value,
	 * all other registers remain the same. */
	for (i = 0; i < 0x80; i++) {
		acx_s_read_phy_reg(priv, i, p++);
	}

	FN_EXIT1(p - buf);
	return p - buf;
}


/***********************************************************************
** acx_e_read_proc_XXXX
** Handle our /proc entry
**
** Arguments:
**	standard kernel read_proc interface
** Returns:
**	number of bytes written to buf
** Side effects:
**	none
*/
static int
acx_e_read_proc(char *buf, char **start, off_t offset, int count,
		     int *eof, void *data)
{
	wlandevice_t *priv = (wlandevice_t *)data;
	unsigned long flags;
	int length;

	FN_ENTER;

	acx_sem_lock(priv);
	acx_lock(priv, flags);
	/* fill buf */
	length = acx_l_proc_output(buf, priv);
	acx_unlock(priv, flags);
	acx_sem_unlock(priv);

	/* housekeeping */
	if (length <= offset + count)
		*eof = 1;
	*start = buf + offset;
	length -= offset;
	if (length > count)
		length = count;
	if (length < 0)
		length = 0;
	FN_EXIT1(length);
	return length;
}

static int
acx_e_read_proc_diag(char *buf, char **start, off_t offset, int count,
		     int *eof, void *data)
{
	wlandevice_t *priv = (wlandevice_t *)data;
	int length;

	FN_ENTER;

	acx_sem_lock(priv);
	/* fill buf */
	length = acx_s_proc_diag_output(buf, priv);
	acx_sem_unlock(priv);

	/* housekeeping */
	if (length <= offset + count)
		*eof = 1;
	*start = buf + offset;
	length -= offset;
	if (length > count)
		length = count;
	if (length < 0)
		length = 0;
	FN_EXIT1(length);
	return length;
}

static int
acx_e_read_proc_eeprom(char *buf, char **start, off_t offset, int count,
		     int *eof, void *data)
{
	wlandevice_t *priv = (wlandevice_t *)data;
	int length;

	FN_ENTER;

	/* fill buf */
	length = 0;
	if (IS_PCI(priv)) {
		acx_sem_lock(priv);
		length = acx_proc_eeprom_output(buf, priv);
		acx_sem_unlock(priv);
	}

	/* housekeeping */
	if (length <= offset + count)
		*eof = 1;
	*start = buf + offset;
	length -= offset;
	if (length > count)
		length = count;
	if (length < 0)
		length = 0;
	FN_EXIT1(length);
	return length;
}

static int
acx_e_read_proc_phy(char *buf, char **start, off_t offset, int count,
		     int *eof, void *data)
{
	wlandevice_t *priv = (wlandevice_t *)data;
	int length;

	FN_ENTER;

	acx_sem_lock(priv);
	/* fill buf */
	length = acx_s_proc_phy_output(buf, priv);
	acx_sem_unlock(priv);

	/* housekeeping */
	if (length <= offset + count)
		*eof = 1;
	*start = buf + offset;
	length -= offset;
	if (length > count)
		length = count;
	if (length < 0)
		length = 0;
	FN_EXIT1(length);
	return length;
}


/***********************************************************************
** /proc files registration
*/
static const char * const
proc_files[] = { "", "_diag", "_eeprom", "_phy" };

static read_proc_t * const
acx_proc_funcs[] = {
	acx_e_read_proc,
	acx_e_read_proc_diag,
	acx_e_read_proc_eeprom,
	acx_e_read_proc_phy
};

static int
manage_proc_entries(const struct net_device *dev, int remove)
{
	/* doh, netdev_priv() doesn't have const! */
	wlandevice_t *priv = acx_netdev_priv((struct net_device *)dev);
	char procbuf[80];
	int i;

	for (i = 0; i < 4; i++)	{
		sprintf(procbuf, "driver/acx_%s", dev->name);
		strcat(procbuf, proc_files[i]);
		if (!remove) {
			acxlog(L_INIT, "creating /proc entry %s\n", procbuf);
			if (!create_proc_read_entry(procbuf, 0, 0, acx_proc_funcs[i], priv))
				return NOT_OK;
		} else {
			acxlog(L_INIT, "removing /proc entry %s\n", procbuf);
			remove_proc_entry(procbuf, NULL);
		}
	}
	return OK;
}

int
acx_proc_register_entries(const struct net_device *dev)
{
	return manage_proc_entries(dev, 0);
}

int
acx_proc_unregister_entries(const struct net_device *dev)
{
	return manage_proc_entries(dev, 1);
}
#endif /* CONFIG_PROC_FS */


/***********************************************************************
** acx_s_set_defaults
** Called from acx_s_init_mac
*/
int
acx_s_set_defaults(wlandevice_t *priv)
{
	unsigned long flags;

	FN_ENTER;

	/* query some settings from the card.
	 * NOTE: for some settings, e.g. CCA and ED (ACX100!), an initial
	 * query is REQUIRED, otherwise the card won't work correctly!! */
	priv->get_mask = GETSET_ANTENNA|GETSET_SENSITIVITY|GETSET_STATION_ID|GETSET_REG_DOMAIN;
	/* Only ACX100 supports ED and CCA */
	if (IS_ACX100(priv))
		priv->get_mask |= GETSET_CCA|GETSET_ED_THRESH;

	acx_s_update_card_settings(priv, 0, 0);

	acx_lock(priv, flags);

	/* set our global interrupt mask */
	if (IS_PCI(priv))
		acx_set_interrupt_mask(priv);

	priv->led_power = 1; /* LED is active on startup */
	priv->brange_max_quality = 60; /* LED blink max quality is 60 */
	priv->brange_time_last_state_change = jiffies;

	/* copy the MAC address we just got from the card
	 * into our MAC address used during current 802.11 session */
	MAC_COPY(priv->dev_addr, priv->netdev->dev_addr);
	sprintf(priv->essid, "STA%02X%02X%02X",
		priv->dev_addr[3], priv->dev_addr[4], priv->dev_addr[5]);
	priv->essid_len = sizeof("STAxxxxxx") - 1; /* make sure to adapt if changed above! */
	priv->essid_active = 1;

	/* we have a nick field to waste, so why not abuse it
	 * to announce the driver version? ;-) */
	strncpy(priv->nick, "acx " WLAN_RELEASE, IW_ESSID_MAX_SIZE);

	if (IS_PCI(priv)) {
		if (IS_ACX111(priv)) {
			/* Hope this is correct, only tested with domain 0x30 */
			acx_read_eeprom_offset(priv, 0x16F, &priv->reg_dom_id);
		} else if (priv->eeprom_version < 5) {
			acx_read_eeprom_offset(priv, 0x16F, &priv->reg_dom_id);
		} else {
			acx_read_eeprom_offset(priv, 0x171, &priv->reg_dom_id);
		}
	}

	priv->channel = 1;
	/* 0xffff would be better, but then we won't get a "scan complete"
	 * interrupt, so our current infrastructure will fail: */
	priv->scan_count = 1;
	priv->scan_mode = ACX_SCAN_OPT_PASSIVE;
	/* Doesn't work for acx100, do it only for acx111 for now */
	if (IS_ACX111(priv)) {
		priv->scan_mode = ACX_SCAN_OPT_ACTIVE;
	}
	priv->scan_duration = 100;
	priv->scan_probe_delay = 200;
	priv->scan_rate = ACX_SCAN_RATE_1;

	priv->auth_alg = WLAN_AUTH_ALG_OPENSYSTEM;
	priv->preamble_mode = 2; /* auto */
	priv->listen_interval = 100;
	priv->beacon_interval = DEFAULT_BEACON_INTERVAL;
	priv->mode = ACX_MODE_OFF;
	priv->dtim_interval = DEFAULT_DTIM_INTERVAL;

	priv->msdu_lifetime = DEFAULT_MSDU_LIFETIME;
	SET_BIT(priv->set_mask, SET_MSDU_LIFETIME);

	priv->rts_threshold = DEFAULT_RTS_THRESHOLD;

	/* use standard default values for retry limits */
	priv->short_retry = 7; /* max. retries for (short) non-RTS packets */
	priv->long_retry = 4; /* max. retries for long (RTS) packets */
	SET_BIT(priv->set_mask, GETSET_RETRY);

	priv->fallback_threshold = 3;
	priv->stepup_threshold = 10;
	priv->rate_bcast = RATE111_1;
	priv->rate_bcast100 = RATE100_1;
	priv->rate_basic = RATE111_1 | RATE111_2;
	priv->rate_auto = 1;
	if (IS_ACX111(priv)) {
		priv->rate_oper = RATE111_ALL;
	} else {
		priv->rate_oper = RATE111_ACX100_COMPAT;
	}

	/* configure card to do rate fallback when in auto rate mode. */
	SET_BIT(priv->set_mask, SET_RATE_FALLBACK);

	/* Supported Rates element - the rates here are given in units of
	 * 500 kbit/s, plus 0x80 added. See 802.11-1999.pdf item 7.3.2.2 */
	acx_l_update_ratevector(priv);

	priv->capab_short = 0;
	priv->capab_pbcc = 1;
	priv->capab_agility = 0;

	SET_BIT(priv->set_mask, SET_RXCONFIG);

	/* set some more defaults */
	if (IS_ACX111(priv)) {
		/* 30mW (15dBm) is default, at least in my acx111 card: */
		priv->tx_level_dbm = 15;
	} else {
		/* don't use max. level, since it might be dangerous
		 * (e.g. WRT54G people experience
		 * excessive Tx power damage!) */
		priv->tx_level_dbm = 18;
	}
	/* priv->tx_level_auto = 1; */
	SET_BIT(priv->set_mask, GETSET_TXPOWER);

	if (IS_ACX111(priv)) {
		/* start with sensitivity level 1 out of 3: */
		priv->sensitivity = 1;
	}

	/* better re-init the antenna value we got above */
	SET_BIT(priv->set_mask, GETSET_ANTENNA);

	priv->ps_wakeup_cfg = 0;
	priv->ps_listen_interval = 0;
	priv->ps_options = 0;
	priv->ps_hangover_period = 0;
	priv->ps_enhanced_transition_time = 0;
#ifdef POWER_SAVE_80211
	SET_BIT(priv->set_mask, GETSET_POWER_80211);
#endif

	MAC_BCAST(priv->ap);

	acx_unlock(priv, flags);
	acx_lock_unhold(); // hold time 844814 CPU ticks @2GHz

	acx_s_initialize_rx_config(priv);

	FN_EXIT1(OK);
	return OK;
}


/***********************************************************************
** FIXME: this should be solved in a general way for all radio types
** by decoding the radio firmware module,
** since it probably has some standard structure describing how to
** set the power level of the radio module which it controls.
** Or maybe not, since the radio module probably has a function interface
** instead which then manages Tx level programming :-\
*/
static int
acx111_s_set_tx_level(wlandevice_t *priv, u8 level_dbm)
{
	struct ACX111TxLevel tx_level;

	/* my acx111 card has two power levels in its configoptions (== EEPROM):
	 * 1 (30mW) [15dBm]
	 * 2 (10mW) [10dBm]
	 * For now, just assume all other acx111 cards have the same.
	 * Ideally we would query it here, but we first need a
	 * standard way to query individual configoptions easily. */
	if (level_dbm <= 12) {
		tx_level.level = 2; /* 10 dBm */
		priv->tx_level_dbm = 10;
	} else {
		tx_level.level = 1; /* 15 dBm */
		priv->tx_level_dbm = 15;
	}
	if (level_dbm != priv->tx_level_dbm)
		acxlog(L_INIT, "ACX111 firmware has specific "
			"power levels only: adjusted %d dBm to %d dBm!\n",
			level_dbm, priv->tx_level_dbm);

	return acx_s_configure(priv, &tx_level, ACX1xx_IE_DOT11_TX_POWER_LEVEL);
}

int
acx_s_set_tx_level(wlandevice_t *priv, u8 level_dbm)
{
	if (IS_ACX111(priv)) {
		return acx111_s_set_tx_level(priv, level_dbm);
	}
	if (IS_PCI(priv)) {
		return acx100_s_set_tx_level(priv, level_dbm);
	}
	return OK;
}


/***********************************************************************
*/
#ifdef UNUSED
/* Returns the current tx level (ACX111) */
static u8
acx111_s_get_tx_level(wlandevice_t *priv)
{
	struct ACX111TxLevel tx_level;

	tx_level.level = 0;
	acx_s_interrogate(priv, &tx_level, ACX1xx_IE_DOT11_TX_POWER_LEVEL);
	return tx_level.level;
}
#endif


/***********************************************************************
** acx_s_init_mac
*/
int
acx_s_init_mac(netdevice_t *dev)
{
	wlandevice_t *priv = acx_netdev_priv(dev);
	int result = NOT_OK;

	FN_ENTER;

	if (IS_PCI(priv)) {
		priv->memblocksize = 256; /* 256 is default */
		acx_init_mboxes(priv);
		/* try to load radio for both ACX100 and ACX111, since both
		 * chips have at least some firmware versions making use of an
		 * external radio module */
		acx_s_upload_radio(priv);
	} else {
		priv->memblocksize = 128;
	}

	if (IS_ACX111(priv)) {
		/* for ACX111, the order is different from ACX100
		   1. init packet templates
		   2. create station context and create dma regions
		   3. init wep default keys
		*/
		if (OK != acx111_s_init_packet_templates(priv))
			goto fail;

		if (OK != acx111_s_create_dma_regions(priv)) {
			printk("%s: acx111_create_dma_regions FAILED\n",
							dev->name);
			goto fail;
		}
#ifdef DEBUG_WEP
		/* don't decrypt WEP in firmware */
		if (OK != acx111_s_feature_on(priv, 0, FEATURE2_SNIFFER))
			goto fail;
#endif
	} else {
		if (OK != acx100_s_init_wep(priv))
			goto fail;
		acxlog(L_DEBUG, "between init_wep and init_packet_templates\n");
		if (OK != acx100_s_init_packet_templates(priv))
			goto fail;

		if (OK != acx100_s_create_dma_regions(priv)) {
			printk("%s: acx100_create_dma_regions FAILED\n",
							dev->name);
			goto fail;
		}
	}

	MAC_COPY(dev->dev_addr, priv->dev_addr);
	result = OK;

fail:
	FN_EXIT1(result);
	return result;
}
