/*
  Copyright 2004, 2005 Jean-Baptiste Note

  This file is part of prism54usb.

  prism54usb is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  prism54usb is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with prism54usb; if not, write to the Free Software Foundation,
  Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 
*/

#include <linux/config.h>
#include <linux/kernel.h>
#include <linux/errno.h>
#include <linux/init.h>
#include <linux/slab.h>
#include <linux/module.h>
#include <linux/smp_lock.h>
#include <linux/completion.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/wireless.h>
#include <linux/if_arp.h>
#include <asm/uaccess.h>
#include <linux/usb.h>

#ifdef MADWIFI
/* This is a _bad_ necessity */
#include <linux/version.h>
#include "net80211/if_media.h"
#include "net80211/ieee80211_var.h"
#endif /* MADWIFI */

#include "isl_sm.h"
#include "islsm_protocol.h"
#include "islsm_ioctl.h"
#include "sent_data_ok.h"
//#include "isl_38xx.h"

#define CONFIG_ISLSM_DEBUG
#ifdef CONFIG_ISLSM_DEBUG
int sm_debug = 2;
#else
int sm_debug = 0;
#endif /* CONFIG_ISLSM_DEBUG */

MODULE_DESCRIPTION("Prism54 softmac layer driver");
MODULE_AUTHOR("Jean-Baptiste Note <jean-baptiste.note@m4x.org>");
MODULE_LICENSE("GPL");
MODULE_PARM(sm_debug, "i");
MODULE_PARM_DESC(sm_debug, "Debug enabled or not for softmac stack");

const char dummy_mac[ETH_ALEN] = {0x00, 0x3d, 0xb4, 0x00, 0x00, 0x00};

static int islsm_close(struct net_device *netdev)
{
	struct islsm *islsm = netdev_priv(netdev);
	
	// Is this called in case of a hot unplug ?
	islsm_info("Close function called");	

	// Shutdown should be called unconditionally
	if (netdev->flags & IFF_RUNNING)
	{
		islsm_info("Running on close, shutting down 802.11");
/* this should pertain to stop only */
		// Empty the lmac queues !
		islsm_empty_queues(netdev);
#ifdef MADWIFI
		/* stop the 802.11 ?
		 * Done last because the bh won't run after this */
		/* I'm not even sure i want this ; pump stops the
		 * interface and reboots it... We don't want to look
		 * ieee80211 state here ! */
		ieee80211_new_state(&islsm->sc_ic, IEEE80211_S_INIT, -1);
		islsm_info("Shutdown ieee80211 layer");
#endif
		netif_stop_queue(netdev);
		netdev->flags &= ~IFF_RUNNING;
	}
	return 0;
}

static void islsm_tx_timeout(struct net_device *netdev)
{
	islsm_info("islsm_tx_timeout called");
	return;
}

static int islsm_transmit(struct sk_buff *skb, struct net_device *netdev)
{
	int err = 0;
#ifdef MADWIFI
	struct islsm *islsm = netdev_priv(netdev);
	struct ieee80211com *ic = &islsm->sc_ic;
	struct ieee80211_node *ni = NULL;
	unsigned short padding = 4;

	islsm_info("islsm_transmit called.");

	if ((netdev->flags & IFF_RUNNING) == 0) {
		islsm_dbg("%s: discard, flags %x\n",
			 __func__, netdev->flags);
		return -ENETDOWN;
	}
	
	/*
	 * No data frames go out unless we're associated; this
	 * should not happen as the 802.11 layer does not enable
	 * the xmit queue until we enter the RUN state.
	 */
	if (ic->ic_state != IEEE80211_S_RUN) {
		islsm_dbg("%s: discard, state %u\n",
			__func__, ic->ic_state);
		goto bad;
	}

	/*
	 * We can do the work directly now
	 */

	/*
	 * Encapsulate the packet for transmission.
	 */
	skb = ieee80211_encap(ic, skb, &ni);

	if (skb == NULL) {
		islsm_dbg("%s: discard, encapsulation failure\n", __func__);
		goto bad;
	}
	
	err = islsm_data_tx(netdev, padding,
			   ISLSM_TX_DATA_ID, ISLSM_TX_CONTROL_TYPE_DATAACK, ISLSM_TX_CONTROL_FLAGS_ACK,
			   ISLSM_TX_FRAME_MAGIC4_MGMT,ISLSM_TX_FRAME_MAGIC5_MGMT,
			   ISLSM_FRAME_DATA, skb);
	if (!err)
	{
		netdev->trans_start = jiffies;
		return 0;
	}
	/* FIXME I think in case of a successfull transmit, we must put
	 * a tx value up to date */

bad:
	if (ni && ni != ic->ic_bss)
		ieee80211_free_node(ic, ni);
#endif /* MADWIFI */
	if (skb)
		dev_kfree_skb(skb);
	return err;	/* NB: return !0 only in a ``hard error condition'' */
}

static struct net_device_stats *
islsm_statistics(struct net_device *netdev)
{
	struct islsm *islsm = netdev_priv(netdev);
	/* trigger a statistics update
	   The data collected is not used for now */
	if (netdev->flags & IFF_RUNNING)
		(void) islsm_stats_readback(netdev);
	return &islsm->statistics;
}

#ifdef MADWIFI
/* 
 * Better ioctl stub
 */
static int 
islsm_ioctl(struct net_device *ndev, struct ifreq *rq, int cmd)
{
	struct islsm *islsm = netdev_priv(ndev);
	struct ieee80211com *ic = &islsm->sc_ic;
	int error;

	/* before implementing locking, understand why it's
	 * needed. Access to which structure ? -- stats for instance ?*/

	switch (cmd) {
	case SIOCETHTOOL:
		islsm_info("SIOCETHTOOL called, not supported");
		error = -ENOTSUPP;
		break;
	default:
		error = ieee80211_ioctl(ic, rq, cmd);
		break;
	}
	if (error)
		islsm_info("problem with ioctl %02x, return value %i",cmd,error);
	return error;
}

/* 
 * MADWIFI STACK SPECIFIC CALLBACKS ; dummy for now, i'll put in
 * commentary as to whether they're needed or not
 */
static int
islsm_beacon_alloc(struct net_device* netdev, struct ieee80211_node *ni)
{
	struct islsm *islsm = netdev_priv(netdev);
	struct ieee80211com *ic = &islsm->sc_ic;
	struct sk_buff *skb;
	int err;
	unsigned short padding = 4;
	unsigned int type;

	/* get the frame from the upper layer */
	skb = ieee80211_beacon_alloc(ic, ni, &islsm->islsm_boff);
	if (skb == NULL)
		return -ENOMEM;

	/* those functions free the skb as soon as usb has tx'd it.
	   Are we allowed to do this ? I'd guess so, but should be
	   checked. Jan ? */
	type = islsm->device_version == ISLSM_DEVICE_USB_VER1 ? ISLSM_TX_CONTROL_TYPE_DATANOACK : ISLSM_TX_CONTROL_TYPE_TXVER2;
	err = islsm_data_tx(netdev, padding,
			   ISLSM_TX_BEACON_ID, type, ISLSM_TX_CONTROL_FLAGS_NOACK,
			   0,0,
			   ISLSM_FRAME_BEACON, skb);

	return err;
}

static int
islsm_newstate(struct ieee80211com *ic, enum ieee80211_state nstate, int arg)
{
	struct net_device *netdev = ic->ic_dev;
	struct islsm *islsm = netdev_priv(netdev);

	u8 *bssid;
	u16 mode = ISLSM_TX_CONTROL_FILTER_NOTYPE;
	int error = 0;
	struct ieee80211_node *ni = 0;
	unsigned int channel = 0;
	struct ieee80211_channel *chan = 0;

	u16 chan_m1, chan_m2;
	u16 magic2 = ISLSM_TX_CONTROL_FILTER_MAGIC2_FORM2;
	u32 magic3 = ISLSM_TX_CONTROL_FILTER_MAGIC3_FORM1;
	u32 magic4 = ISLSM_TX_CONTROL_FILTER_MAGIC4_FORM1;
	u16 magic8 = 0;
	u16 magic9 = 0;

	/* what if it has never been initialized, is it safe ? */
	del_timer(&islsm->scan_ch);

	islsm_dbg("%s: %s -> %s\n", __func__,
		 ieee80211_state_name[ic->ic_state],
		 ieee80211_state_name[nstate]);

	netif_stop_queue(netdev);               /* before we do anything else */
	
	if (nstate == IEEE80211_S_INIT) {
		goto done;
	}

	ni = ic->ic_bss;

	/* compute the elements needed for the rx/tx filter */
	/* bssid */
 	if (nstate == IEEE80211_S_SCAN)
	{
		bssid = netdev->broadcast;
		mode = ISLSM_TX_CONTROL_FILTER_NOTYPE;
		chan_m1 = ISLSM_TX_CONTROL_CHANNEL_MAGIC1_SCAN;
		chan_m2 = ISLSM_TX_CONTROL_CHANNEL_MAGIC2_SCAN;
	}
 	else
	{
		bssid = ni->ni_bssid;
		chan_m1 = ISLSM_TX_CONTROL_CHANNEL_MAGIC1_TX;
		chan_m2 = ISLSM_TX_CONTROL_CHANNEL_MAGIC2_TX;
	}

	/* mode  */
	if ( nstate == IEEE80211_S_AUTH || nstate == IEEE80211_S_RUN )
	{
		switch (ic->ic_opmode) 
		{
		case IEEE80211_M_STA: /* infra station */
			magic2 = ISLSM_TX_CONTROL_FILTER_MAGIC2_STA;
			magic3 = ISLSM_TX_CONTROL_FILTER_MAGIC3_STA;
			/* FIXME ?. rate control ? what ? */
			magic4 = ISLSM_TX_CONTROL_FILTER_MAGIC4_STA;
			magic4 = ISLSM_TX_CONTROL_FILTER_MAGIC4_FORM2;
			magic8 = ISLSM_TX_CONTROL_FILTER_MAGIC8_STA;
			mode = ISLSM_TX_CONTROL_FILTER_STA;
			break;
		case IEEE80211_M_HOSTAP:
		case IEEE80211_M_IBSS: /* ad-hoc */
			mode = ISLSM_TX_CONTROL_FILTER_ADHOC;
			break;
		case IEEE80211_M_MONITOR:
			/* correct this -- doesn't work now */
			bssid = netdev->broadcast;
			chan_m1 = ISLSM_TX_CONTROL_CHANNEL_MAGIC1_TX;
			chan_m2 = ISLSM_TX_CONTROL_CHANNEL_MAGIC2_TX;

			magic2 = ISLSM_TX_CONTROL_FILTER_MAGIC2_MONITOR;
			magic3 = ISLSM_TX_CONTROL_FILTER_MAGIC3_MONITOR;
			magic4 = ISLSM_TX_CONTROL_FILTER_MAGIC4_MONITOR;
			mode = ISLSM_TX_CONTROL_FILTER_MONITOR;
			break;
		case IEEE80211_M_AHDEMO:
			/* no idea */
			break;
		}
	}

	/* should also set rate, etc but i don't know yet how to set them */

	chan = ni->ni_chan;
	channel=ieee80211_mhz2ieee(chan->ic_freq,0);

	// FIXME : we will get errors on hot unplug. is this a problem ?
	// This is a problem. Don't know why, but it is. It seems to
	// make madwifi stuck. So silently ignore.
	(void) islsm_freq_change(netdev, channel, chan->ic_freq,
				chan_m1, chan_m2);

	(void) islsm_set_filter(netdev, mode, bssid, 
			       magic2, magic3, magic4, magic8, magic9);

/* what is ni_associd ?
	more or less...
	if (nstate == IEEE80211_S_RUN && ic->ic_opmode == IEEE80211_M_STA)
		ath_hal_setassocid(ah, bssid, ni->ni_associd);
	else
		ath_hal_setassocid(ah, bssid, 0);
	if (ic->ic_flags & IEEE80211_F_PRIVACY) {
		for (i = 0; i < IEEE80211_WEP_NKID; i++)
			if (ath_hal_keyisvalid(ah, i))
				ath_hal_keysetmac(ah, i, bssid);
	}
*/
	if (nstate == IEEE80211_S_RUN) {
		islsm_dbg("%s(RUN): ic_flags=0x%08x iv=%d bssid=%s "
			 "capinfo=0x%04x chan=%d\n"
			 , __func__
			 , ic->ic_flags
			 , ni->ni_intval
			 , ether_sprintf(ni->ni_bssid)
			 , ni->ni_capinfo
			 , ieee80211_chan2ieee(ic, ni->ni_chan));
		
		/*
		 * Allocate and setup the beacon frame for AP or adhoc mode.
		 */
		
		if (ic->ic_opmode == IEEE80211_M_HOSTAP ||
		    ic->ic_opmode == IEEE80211_M_IBSS) {
			error = islsm_beacon_alloc(netdev, ni);
			if (error != 0)
				goto bad;
		}

		/*
		 * No need for this and / or is done in the rx/tx filter packet
		 * Configure the beacon and sleep timers.
		 */
/*		ath_beacon_config(sc); */
	}
done:
	/*
	 * Invoke the parent method to complete the work.
	 */
	error = islsm->ieee80211_newstate(ic, nstate, arg);

	if (nstate == IEEE80211_S_SCAN) {
		/* start ap/neighbor scan timer */
#define ISLSM_DWELLTIME 1000 /* ms */
 		mod_timer(&islsm->scan_ch,
 			  jiffies + ((HZ * ISLSM_DWELLTIME) / 1000));
		
#undef ISLSM_DWELLTIME
	}
bad:
	netif_start_queue(netdev);
	return error;
}

static void
islsm_next_scan(unsigned long arg)
{
	struct net_device *netdev = (struct net_device *) arg;
	struct islsm *islsm = netdev_priv(netdev);
	struct ieee80211com *ic = &islsm->sc_ic;

	if (ic->ic_state == IEEE80211_S_SCAN)
		ieee80211_next_scan(ic);
}

static int 
islsm_init(struct net_device *netdev) {
	struct islsm *islsm = netdev_priv(netdev);
	struct ieee80211com *ic = &islsm->sc_ic;
	struct ieee80211_node *ni;
	int err = 0;

	u8* bssid = netdev->broadcast;

	// Should be the open of the device -- it's not, yet
	islsm_info("islsm_init called");

	/* resize skbs in monitor mode -- good idea, leave enough
	 * headroom to fit the prism header */

	/* tell the powers that be what we really are */
	netdev->type = (ic->ic_opmode == IEEE80211_M_MONITOR) ?
		ARPHRD_IEEE80211_PRISM : 
		ARPHRD_ETHER;

	/* Hardware state machine reset */
	if (islsm->isl_boot(netdev))
	{
		islsm_info("unable to boot device");
		err = -EIO;
		goto done;
	}

	/* submit initial data on data pipe */
	/* this seems to allow frequency change successfully */
   	islsm_outofband_msg(netdev, 
			    magicpkt_table[islsm->device_version].data,
			    magicpkt_table[islsm->device_version].len, islsm->config_reg);

	islsm_set_filter(netdev,
			 ISLSM_TX_CONTROL_FILTER_NOTYPE, bssid,
			 ISLSM_TX_CONTROL_FILTER_MAGIC2_FORM3, 1, 0, 
			 0, ISLSM_TX_CONTROL_FILTER_MAGIC9_FORM1);

	// This one may not be necessary any more.
	islsm_set_filter(netdev,
			 ISLSM_TX_CONTROL_FILTER_NOTYPE, bssid,
			 ISLSM_TX_CONTROL_FILTER_MAGIC2_FORM2,ISLSM_TX_CONTROL_FILTER_MAGIC3_FORM1,ISLSM_TX_CONTROL_FILTER_MAGIC4_FORM1,
			 0,0);

	// Start the statistics update cycle
//	mod_timer(&islsm->stats_update, jiffies + ((HZ * ISLSM_STATSTIME) / 1000));

	/* Tell others we're okay */
	netdev->flags |= IFF_RUNNING;
	ic->ic_state = IEEE80211_S_INIT;

	/*
	 * The hardware should be ready to go now so it's safe
	 * to kick the 802.11 state machine as it's likely to
	 * immediately call back to us to send mgmt frames.
	 */
	ni = ic->ic_bss;
	ni->ni_chan = ic->ic_ibss_chan;

	// This will set the tx/rx filter ! this is what we need.
	//mode = ieee80211_chan2mode(ic, ni->ni_chan);
	//if (mode != sc->sc_curmode)
	//	ath_setcurmode(sc, mode);
	if (ic->ic_opmode != IEEE80211_M_MONITOR) {
		if (ic->ic_roaming != IEEE80211_ROAMING_MANUAL)
			ieee80211_new_state(ic, IEEE80211_S_SCAN, -1);
	} else
		ieee80211_new_state(ic, IEEE80211_S_RUN, -1);
 done:
	return err;	
}

static void 
islsm_setup_freqs(struct ieee80211com *ic) {
	/* 
	 * FIXME: Should be read from the device, really, but needs complete
	 * initialization, so we don't do this for now
	 * in the _real_ driver : init on probe seems necessary.
	 */
	unsigned int ieee_num, maxchans;
	maxchans = ISLSM_NR_CHANNELS;
	if (maxchans > IEEE80211_CHAN_MAX)
	{
		islsm_dbg("Channel table table too small");
		maxchans = IEEE80211_CHAN_MAX;
	}

 	for(ieee_num = 1 ; ieee_num <= ISLSM_NR_CHANNELS ; ieee_num++)
	{
		ic->ic_channels[ieee_num].ic_freq = ieee80211_ieee2mhz(ieee_num,IEEE80211_CHAN_2GHZ);
		ic->ic_channels[ieee_num].ic_flags = 
			IEEE80211_MODE_11B |
			IEEE80211_MODE_11G;
		islsm_dbg("setup channel ieee %i, freq %iMHz",
			 ieee_num, ic->ic_channels[ieee_num].ic_freq);
	}
}

/* needed otherwise problems in */
static int
islsm_setup_rates(struct net_device *dev, u_int mode)
{
	struct islsm *islsm = dev->priv;
	struct ieee80211com *ic = &islsm->sc_ic;
	struct ieee80211_rateset *rs;
	int i, maxrates;
	/* standard rates observed from my prism54 in hostap: 
	   
	2, basic yes rate 4, basic yes rate 11, basic yes rate 12, basic
	   no rate 18, basic no rate 22, basic yes rate 24, basic no
	   rate 36, basic no rate 48, basic no rate 72, basic no rate
	   96, basic no rate 108,
	*/
	int rates[ISLSM_NR_RATES] = { 2, 4, 11, 12, 18, 22, 24, 36, 48, 72, 96, 108, };

	/* FIXME: should check mode */
	rs = &ic->ic_sup_rates[mode];
	/* FIXME: Should also get them from device ! */
	maxrates = ISLSM_NR_RATES;
	if (maxrates > IEEE80211_RATE_MAXSIZE)
	{
		islsm_info("Rate table too small");
		maxrates = IEEE80211_RATE_MAXSIZE;
	}
	for (i = 0; i < maxrates; i++)
#define MADWIFI_RATE(x) ( (x) & IEEE80211_RATE_VAL )
		rs->rs_rates[i] = MADWIFI_RATE(rates[i]);
#undef MADWIFI_RATE
	rs->rs_nrates = maxrates;
	return 0;
}

/*
 * Transmit a management frame.  On failure we reclaim the skbuff.
 * Note that management frames come directly from the 802.11 layer
 * and do not honor the send queue flow control.  Need to investigate
 * using priority queueing so management frames can bypass data.
 * Does nothing for now.
 */

/*
 * What i know-how : alloc a frame, then send it during _active_ _scan_.
 * I don't know how to manage this in this framework...
 */ 
static int 
islsm_mgtstart(struct ieee80211com *ic, struct sk_buff *skb) {
	/* we do not free the skb here, lower-level functions do this */
	struct net_device *netdev = ic->ic_dev;
	struct islsm *islsm = netdev_priv(netdev);
	int err=0;
	unsigned short padding;

	islsm_info("islsm_mgtstart called");

	/* only probe */
	if (ic->ic_state == IEEE80211_S_SCAN) {
		struct ieee80211_node *ni = 0;
		struct ieee80211_channel *chan = 0;
		unsigned int channel = 0;
		unsigned int type;
		/* then i consider that we're active scanning
		   This is a horrible kludge. What we'd need would be 
		   a probe alloc similar to the beacon alloc */

		padding = 2;
		type = islsm->device_version == ISLSM_DEVICE_USB_VER1 ? 
			ISLSM_TX_CONTROL_TYPE_DATANOACK : 
			ISLSM_TX_CONTROL_TYPE_TXVER2;
		err = islsm_data_tx(netdev, padding,
				   ISLSM_TX_MGMT_ID, type, ISLSM_TX_CONTROL_FLAGS_NOACK,
				   0,0,
				   ISLSM_FRAME_PROBE, skb);
		if (err)
			return err;
		
		/* then do the frequency change required to emit it */
		ni = ic->ic_bss;
		chan = ni->ni_chan;
		channel=ieee80211_mhz2ieee(chan->ic_freq,0);
		err = islsm_freq_change(netdev, channel, chan->ic_freq,
				       ISLSM_TX_CONTROL_CHANNEL_MAGIC1_EMIT_TXSLOT,
				       ISLSM_TX_CONTROL_CHANNEL_MAGIC2_EMIT_TXSLOT);
		goto tx_done;
	}

	else
	{
		/* this is untested */
		padding = 4;
		/* try and see what magic4 and magic5 really mean */
		/* do they mean that the tx must be done ASAP ? */
		err = islsm_data_tx(netdev, padding,
				   ISLSM_TX_MGMT_ID, ISLSM_TX_CONTROL_TYPE_DATAACK, ISLSM_TX_CONTROL_FLAGS_ACK,
				   ISLSM_TX_FRAME_MAGIC4_MGMT,ISLSM_TX_FRAME_MAGIC5_MGMT,
				   ISLSM_FRAME_MGMT, skb);
	}

 tx_done:
	return err;
}

/*
 * Reset the hardware w/o losing operational state.  This is
 * basically a more efficient way of doing ath_stop, ath_init,
 * followed by state transitions to the current 802.11
 * operational state.  Used to recover from errors rx overrun
 * and to reset the hardware when rf gain settings must be reset.
 */
static int
islsm_reset(struct net_device *netdev) {
	struct islsm *islsm = netdev_priv(netdev);
	struct ieee80211com *ic = &islsm->sc_ic;
	// What's the difference with ic->ni->... I don't know but i'll
	// have to figure out for Ad-Hoc mode i guess
	struct ieee80211_channel *chan = ic->ic_ibss_chan;
	unsigned int channel = 0;

	// Dirty code, should factor with newstate and mgtstart into a frequency
	// change function that does all things needed...
	u16 chan_m1 = ISLSM_TX_CONTROL_CHANNEL_MAGIC1_TX;
	u16 chan_m2 = ISLSM_TX_CONTROL_CHANNEL_MAGIC2_TX;

	islsm_info("call to islsm_reset, but only changing frequency");

	channel=ieee80211_mhz2ieee(chan->ic_freq,0);
	islsm_freq_change(netdev, channel, chan->ic_freq,
			 chan_m1, chan_m2);

	if (ic->ic_state == IEEE80211_S_RUN)
		netif_wake_queue(netdev);	/* restart xmit */

	return 0;
}

/*
 * Setup driver-specific state for a newly associated node.
 * Note that we're called also on a re-associate, the isnew
 * param tells us if this is the first time or not.
 */
static void
islsm_newassoc(struct ieee80211com *ic, struct ieee80211_node *ni, int isnew)
{
	(void) ic;
	(void) ni;
	(void) isnew;
	islsm_info("call to islsm_newassoc");
	return;
}


/*
 * Callback from the 802.11 layer to update the
 * slot time based on the current setting.
 */
static void
islsm_updateslot(struct net_device *dev)
{
	(void) dev;
	islsm_info("call to islsm_updateslot");
	return;
}

static int
islsm_media_change(struct net_device *dev)
{
	int error;
	islsm_info("call to islsm_media_change");

	error = ieee80211_media_change(dev);
	if (error == ENETRESET) {
		if ((dev->flags & (IFF_RUNNING|IFF_UP)) == (IFF_RUNNING|IFF_UP))
			error = islsm_init(dev);
		else
			error = 0;
	}
	return error;
}

#else /* MADWIFI */
static int 
islsm_ioctl(struct net_device *ndev, struct ifreq *rq, int cmd)
{
	islsm_info("islsm_ioctl called returning not supported");
	return -ENOTSUPP;
}

static int
islsm_init(struct net_device *netdev) {
	struct islsm *islsm = netdev_priv(netdev);
	int err = 0;

	u8* bssid = netdev->broadcast;

	islsm_info("islsm_init called");	

	/* resize skbs in monitor mode -- good idea, leave enough
	 * headroom to fit the prism header */

	/* Hardware state machine reset */
	if (islsm->isl_boot(netdev))
	{
		islsm_info("unable to boot device");
		err = -EIO;
		goto done;
	}

	/* submit initial data on data pipe */
	/* this seems to allow frequency change successfully */
   	islsm_outofband_msg(netdev, 
			    magicpkt_table[islsm->device_version].data,
			    magicpkt_table[islsm->device_version].len,
			    islsm->config_reg);

	islsm_set_filter(netdev,
			 ISLSM_TX_CONTROL_FILTER_NOTYPE, bssid,
			 ISLSM_TX_CONTROL_FILTER_MAGIC2_FORM3, 1, 0, 
			 0, ISLSM_TX_CONTROL_FILTER_MAGIC9_FORM1);

	// This one may not be necessary any more.
	islsm_set_filter(netdev,
			 ISLSM_TX_CONTROL_FILTER_NOTYPE, bssid,
			 ISLSM_TX_CONTROL_FILTER_MAGIC2_FORM2,ISLSM_TX_CONTROL_FILTER_MAGIC3_FORM1,ISLSM_TX_CONTROL_FILTER_MAGIC4_FORM1,
			0,0);

	// Start the statistics update cycle
//	mod_timer(&islsm->stats_update, jiffies + ((HZ * ISLSM_STATSTIME) / 1000));

	/* Tell others we're okay */
	netdev->flags |= IFF_RUNNING;
	netif_start_queue(netdev);
	
 done:
	return err;
}

#endif /* MADWIFI */

int 
islsm_setup_net(struct net_device *netdev)
{
	struct islsm *islsm = netdev_priv(netdev);
#ifdef MADWIFI
	struct ieee80211com *ic = &islsm->sc_ic;
#endif /* MADWIFI */
	int error;
	/* initialize the function pointers */
	/* FIXME : which ones are redefined, ifdef them */
	netdev->open = &islsm_init;
	netdev->stop = &islsm_close;
	netdev->get_stats = &islsm_statistics;
	netdev->do_ioctl = &islsm_ioctl;
	netdev->hard_start_xmit = &islsm_transmit;
	netdev->addr_len = ETH_ALEN;
	// FIXME : check it has not been set / overwritten by the madwifi...
	// FIXME : madwifi doesn't care about this, so we're screwed
	netdev->hard_header_len = FULL_TX_CONTROL_ALLOCDATA + TX_MAX_PADDING + islsm->device_tx_header_space;
	memcpy(netdev->dev_addr, dummy_mac, ETH_ALEN);
//	netdev->set_mac_address = &prism54_set_mac_address;
	netdev->watchdog_timeo = ISLSM_TX_TIMEOUT;
	netdev->tx_timeout = &islsm_tx_timeout;

#ifdef CONFIG_NET_WIRELESS
	netdev->get_wireless_stats = &islsm_wireless_stats;
#ifdef MADWIFI
	islsm_dbg("Madwifi ioctl-control");
	ieee80211_ioctl_iwsetup(&islsm_iw_handler_def);
#endif /* MADWIFI */
	/* check if this can be done right away. Race ? */
	netdev->wireless_handlers = (struct iw_handler_def *) &islsm_iw_handler_def;
#endif /* CONFIG_NET_WIRELESS */

#ifdef MADWIFI 
	islsm_info("Start madwifi dev configuration");
	/* copied verbatim for now, see if it works */
        /* ic variables initialization */
	ic->ic_dev = netdev;
	/* verbosity of the stack : full ! */
	ic->msg_enable = (u_int32_t) -1;

	/* The net_device_stats struct associated with your device */
	ic->ic_devstats = &islsm->statistics;
	/* callbacks : see above */
	ic->ic_mgtstart = islsm_mgtstart;
	/* init callback : called on frequency change */
	ic->ic_init = islsm_init;
	ic->ic_reset = islsm_reset;
	ic->ic_newassoc = islsm_newassoc;
	ic->ic_updateslot = islsm_updateslot;

	/* capabilities */
	/* frequencies */
	islsm_setup_freqs(ic);
	islsm_setup_rates(netdev, IEEE80211_MODE_11B);
	islsm_setup_rates(netdev, IEEE80211_MODE_11G);

	/* XXX not right but it's not used anywhere important */
	ic->ic_phytype = IEEE80211_T_OFDM;
	ic->ic_opmode = IEEE80211_M_STA;
	ic->ic_caps =
		  IEEE80211_C_IBSS		/* ibss, nee adhoc, mode */
		| IEEE80211_C_HOSTAP		/* hostap mode */
		| IEEE80211_C_MONITOR		/* monitor mode */
		| IEEE80211_C_SHPREAMBLE	/* short preamble supported */
		| IEEE80211_C_SHSLOT		/* short slot time supported */
		| IEEE80211_C_TXPMGT		/* transmit power control */
		;

	/*
	 * Indicate we need the 802.11 header padded to a
	 * 32-bit boundary for 4-address and QoS frames.
	 */
	ic->ic_flags |= IEEE80211_F_DATAPAD;
	
	/* to be done elsewhere */
	/* 
	   struct *sc = netdev_priv(netdev);
	   struct ieee80211com *ic = &sc->sc_ic;
	   IEEE80211_ADDR_COPY(dev->dev_addr, ic->ic_myaddr);
	 */
	/* call MI attach routine. */

	init_timer(&islsm->scan_ch);
	islsm->scan_ch.function = islsm_next_scan;
	islsm->scan_ch.data = (unsigned long) netdev;

/* 	init_timer(&islsm->stats_update); */
/* 	islsm->stats_update.function = islsm_stats_update; */
/* 	islsm->stats_update.data = (unsigned long) netdev; */

	islsm_dbg("Start madwifi dev attach");
	ieee80211_ifattach(ic);
	islsm_dbg("madwifi dev attach done.\n");

	/* hooks into default methods */
	islsm->ieee80211_newstate = ic->ic_newstate;
	ic->ic_newstate = islsm_newstate;

	/* complete initialization */
	islsm_dbg("Call madwifi media init");
	ieee80211_media_init(ic, islsm_media_change, ieee80211_media_status);
	islsm_dbg("madwifi media init done.\n");

#endif /* MADWIFI */

	error=register_netdev(netdev);
	if (error) {
		islsm_info("%s: unable to register device\n", netdev->name);
		goto out;
	}
	islsm_info("Prism54 SoftMac device now attached to %s", netdev->name);

#ifdef MADWIFI
#ifdef CONFIG_SYSCTL
	ieee80211_sysctl_register(ic);
#endif /* CONFIG_SYSCTL */
	islsm_dbg("Call madwifi announce");
	ieee80211_announce(ic);
	islsm_dbg("madwifi announce done.\n");
#endif /* MADWIFI */

	/* LMAC initialization -- may be moved after parse mgmt frame at
	 * some point in time... */
	islsm_txqueue_init(netdev);
	init_completion(&islsm->dev_init_comp);

	return 0;
	
 out:
#ifdef MADWIFI
	ieee80211_ifdetach(ic);
#endif
	return error;
}
EXPORT_SYMBOL(islsm_setup_net);

int 
islsm_unsetup_net(struct net_device *netdev)
{
	struct islsm *islsm = netdev_priv(netdev);

	/* shut down the device */
	islsm_close(netdev);

//	del_timer_sync(&islsm->stats_update);

#ifdef MADWIFI
	islsm_dbg("Detaching madwifi");
	ieee80211_ifdetach(&islsm->sc_ic);
	islsm_dbg("Madwifi detached");
#endif
	unregister_netdev(netdev);
	islsm_dbg("Unregistered netdevice");
	return 0;
}
EXPORT_SYMBOL(islsm_unsetup_net);

static int __init islsm_modinit(void)
{
	return 0;
}

static void __exit islsm_modexit(void)
{
	return;
}

module_init(islsm_modinit);
module_exit(islsm_modexit);
