/*
	rt2x00 SourceForge Project - <http://rt2x00.serialmonkey.com>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the
	Free Software Foundation, Inc.,
	59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

	Licensed under the GNU GPL
	Original code supplied under license from RaLink Inc, 2003
 */

/*
	Module Name: rt2x00_interrupt.c

	Abstract: All routines to required to send and receive packets.

	Revision History:
	Who		When		What
	--------	-----------	-----------------------------
	IvD		<releasedate>	Initial Release.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>

#include <net/iw_handler.h>
#include <ieee802_11.h>

#include "rt2x00_mgmt.h"
#include "rt2x00_adapter.h"

/* 
 * See IEEE 802.1H for LLC/SNAP encapsulation/decapsulation
 *  Ethernet-II snap header (RFC1042) - most EtherTypes
 *  Bridge-Tunnel header - EtherTypes ETH_P_AARP and ETH_P_IPX
 *  No encapsulation header - EtherType < 0x600 (=length)
 */
static u8 rfc1042_header[ETH_ALEN] = { 0xaa, 0xaa, 0x03, 0x00, 0x00, 0x00 };
static u8 bridge_tunnel_header[ETH_ALEN] = { 0xaa, 0xaa, 0x03, 0x00, 0x00, 0xf8 };
#define EAPOL		0x888e	/* EAPOL - 802.11x protocol*/

/*
 * Almost all packets require the headers to be changed.
 * Incoming packets require the 802.11 headers to be replaced with 802.3,
 * and packets to be transmitted require the 802.3 header to be replaced with the 802.11 version.
 */
struct sk_buff* rt2x00_construct_802_11hdr(struct _rt2x00_adapter *adapter, struct sk_buff *skb)
{
	struct sk_buff		*tmp;
	struct ethhdr		*eth_header = eth_hdr(skb);
	struct ieee802_11_hdr	header802_11;

	memset(&header802_11, 0x00, sizeof(struct ieee802_11_hdr));

	/*
	 * Check for 802.11x protocol (EAPOL).
	 * On a non-WPA network, 802.11x packets should never be encrypted.
	 */
	if(TEST_FLAG(adapter, CRYPTO_WEP) && eth_header->h_proto != EAPOL)
		header802_11.frame_ctl |= IEEE802_11_FCTL_WEP;

	if(TEST_FLAG(adapter, ENABLED_POWER))
		header802_11.frame_ctl |= IEEE802_11_FCTL_PM;

	header802_11.frame_ctl |= IEEE802_11_FTYPE_DATA;				/* Frame type. */

	if(LINK_TYPE(adapter, IW_MODE_ADHOC)){
		memcpy(header802_11.addr1, eth_header->h_dest, ETH_ALEN);		/* Address 1 - Destination Address */
		memcpy(header802_11.addr2, eth_header->h_source, ETH_ALEN);		/* Address 2 - Source Address. */
		memcpy(header802_11.addr3, adapter->config->bssid, ETH_ALEN);		/* Address 3 - BSSID */
	} else if(LINK_TYPE(adapter, IW_MODE_INFRA)) {
		memcpy(header802_11.addr1, adapter->config->bssid, ETH_ALEN);		/* Address 1 - BSSID */
		memcpy(header802_11.addr2, eth_header->h_source, ETH_ALEN);		/* Address 2 - Source Address. */
		memcpy(header802_11.addr3, eth_header->h_dest, ETH_ALEN);		/* Address 3 - Destination Address */
		header802_11.frame_ctl |= IEEE802_11_FCTL_TODS;				/* All packets are send to the AP. */
	} else if(LINK_TYPE(adapter, IW_MODE_MASTER)) {
		memcpy(header802_11.addr1, eth_header->h_dest, ETH_ALEN);		/* Address 1 - Destination Address */
		memcpy(header802_11.addr2, adapter->net_dev->dev_addr, ETH_ALEN);	/* Address 2 - BSSID */
		memcpy(header802_11.addr3, eth_header->h_source, ETH_ALEN);		/* Address 3 - Source Address. */
		header802_11.frame_ctl |= IEEE802_11_FCTL_FROMDS;			/* All packets come from the AP. */
	}

	((struct _skb_cb_rx*)skb->cb)->header_length = IEEE802_11_HLEN; /* Additional header space required in the buffer. */

	if(eth_header->h_proto == ETH_P_AARP || eth_header->h_proto == ETH_P_IPX)
		memcpy(&header802_11.addr4, bridge_tunnel_header, ETH_ALEN);
	else if(eth_header->h_proto >= 0x600)
		memcpy(&header802_11.addr4, rfc1042_header, ETH_ALEN);
	else
		((struct _skb_cb_rx*)skb->cb)->header_length -= ETH_ALEN;


	skb_pull(skb, ETH_HLEN); /* The old ethernet header is no longer required, remove it from the buffer. */

	/* 
	 * We use a temporary pointer to sk_buff, when the reallocation for the headroom fails
	 * we are still able to delete the buffer, and thus prevent memory leaking.
	 */
	tmp = skb;
	skb = skb_copy_expand(skb, sizeof(struct ieee802_11_hdr), 0, GFP_ATOMIC);
	dev_kfree_skb_irq(tmp);
	if(!skb)
		return NULL;

	memset(skb->data, 0x00, sizeof(struct ieee802_11_hdr));
	memcpy(skb->data, &header802_11, sizeof(struct ieee802_11_hdr));

	return skb;
}

struct sk_buff* rt2x00_update_802_11hdr(struct _rt2x00_adapter *adapter, struct sk_buff *skb)
{
	struct sk_buff		*tmp;
	struct ieee802_11_hdr	*header802_11 = (struct ieee802_11_hdr*)skb->data;

	if(!LINK_TYPE(adapter, IW_MODE_SECOND))
		return skb;

	header802_11->frame_ctl |= IEEE802_11_FCTL_FROMDS | IEEE802_11_FCTL_TODS;

	/*
	 * We must place our address in the header to inform the AP
	 * for whom this packet is for that it is updated by us.
	 */
	tmp = skb;
	skb = skb_copy_expand(skb, ETH_ALEN, 0, GFP_ATOMIC);
	dev_kfree_skb_irq(tmp);
	if(!skb)
		return NULL;

	/* 
	 * Update our header pointer.
	 */
	header802_11 = (struct ieee802_11_hdr*)skb->data;

	/*
	 * We need to place our address in addr4 of the 802.11 structure.
	 * This address is located in the last bytes of the header.
	 * But when the addr4 is unsued it is filled with other data.
	 * So move the header minus 1 address length to the beginning of the packet.
	 * And write our address in the address length that has been freed with this action.
	 */
	memmove(header802_11, header802_11 + ETH_ALEN, sizeof(struct ieee802_11_hdr) - ETH_ALEN);
	memcpy(header802_11->addr4, adapter->config->bssid, ETH_ALEN);

	return skb;
}

void rt2x00_construct_802_3hdr(struct _rt2x00_adapter *adapter, struct sk_buff *skb)
{
	struct ethhdr		eth_header;
	struct ieee802_11_hdr	*header802_11 = (struct ieee802_11_hdr*)skb->data;
	struct _skb_cb_rx	*skb_cb = (struct _skb_cb_rx*)skb->cb;

	memset(&eth_header, 0x00, sizeof(struct ethhdr));

	if(LINK_TYPE(adapter, IW_MODE_ADHOC)){
		memcpy(eth_header.h_dest, header802_11->addr1, ETH_ALEN);		/* Address 1 - Destination Address */
		memcpy(eth_header.h_source, header802_11->addr2, ETH_ALEN);		/* Address 2 - Source Address. */
	} else if(LINK_TYPE(adapter, IW_MODE_INFRA)) {
		memcpy(eth_header.h_source, header802_11->addr2, ETH_ALEN);		/* Address 2 - Source Address. */
		memcpy(eth_header.h_dest, header802_11->addr3, ETH_ALEN);		/* Address 3 - Destination Address */
	} else if(LINK_TYPE(adapter, IW_MODE_MASTER)){
		memcpy(eth_header.h_dest, header802_11->addr1, ETH_ALEN);		/* Address 1 - Destination Address */
		memcpy(eth_header.h_source, header802_11->addr3, ETH_ALEN);		/* Address 3 - Source Address. */
	}

	eth_header.h_proto = htons(skb_cb->rxd.data_byte_count);

	/*
	 * We can now replace the old 802.11 header with the newly created 802.3 header.
	 * The 802.11 header is larger then the 802.3 header. This means we will have enough memory
	 * in the skb buffer to add the new header once the old one is removed. 
	 */
	skb_pull(skb, IEEE802_11_HLEN);
	memcpy(skb_push(skb, ETH_HLEN), &eth_header, ETH_HLEN);
}

static inline char rt2x00_filter_validate(struct _rt2x00_adapter *adapter, struct sk_buff *skb)
{
	struct ieee802_11_hdr	*header802_11 = (struct ieee802_11_hdr*)skb->data;
	struct _skb_cb_rx	*skb_cb = (struct _skb_cb_rx*)skb->cb;

	if(skb_cb->rxd.error || skb_cb->rxd.crc){
		adapter->stats.rx_crc_errors++;
		return -EINVAL;
	}

	if(header802_11->frame_ctl & IEEE802_11_FCTL_MOREFRAGS && (skb_cb->rxd.multicast || skb_cb->rxd.broadcast))
		return -EINVAL;

	if(LINK_TYPE(adapter, IW_MODE_MASTER) || LINK_TYPE(adapter, IW_MODE_SECOND)){
		if(!(header802_11->frame_ctl & IEEE802_11_FCTL_TODS))
			return -EINVAL;
	}

	/* beacon packets must be excluded from FROMDS tests */
	if(LINK_TYPE(adapter, IW_MODE_INFRA)){
		if(!(header802_11->frame_ctl & IEEE802_11_STYPE_BEACON))
		    if(!(header802_11->frame_ctl & IEEE802_11_FCTL_FROMDS))
			return -EINVAL;
	}

        return 0;
}

static inline char rt2x00_filter_duplicate(struct _rt2x00_adapter *adapter, struct sk_buff *skb)
{
	struct ieee802_11_hdr	*header802_11 = (struct ieee802_11_hdr*)skb->data;
	u8			counter;

	/*
	 * If this packet is not repeated by the sender, then we do not have it in cache.
	 * Search for an empty entry, and add the new entry.
	 */
	if(!(header802_11->frame_ctl & IEEE802_11_FCTL_RETRY)){
		for(counter = 0; counter < MAX_DUPLICATE_CACHE; counter++)
			if(!adapter->duplicate_cache[counter].sequence)
				goto add_entry;
	}

	for(counter = 0; counter < MAX_DUPLICATE_CACHE; counter++){
		if(!adapter->duplicate_cache[counter].sequence)
			goto add_entry;

		if(memcmp(&header802_11->addr2, &adapter->duplicate_cache[counter].addr, ETH_ALEN) == 0){
			if(adapter->duplicate_cache[counter].sequence == header802_11->seq_ctl)
				return -EINVAL;
			else
				goto update;
		}
	}

	/*
	 * We have reached the end of the cache list, our packet is not a duplicate,
	 * unfortunatly we do not have enough space to store the packet in the cache for future duplicates.
	 */
	return 0;

add_entry: /* Add new entry */
	memcpy(&adapter->duplicate_cache[counter].addr, &header802_11->addr2, ETH_ALEN);
update: /* Update old entry */
	adapter->duplicate_cache[counter].sequence = header802_11->seq_ctl;
	return 0;
}

static inline char rt2x00_filter_multicast(struct _rt2x00_adapter *adapter, char *addr)
{
	struct dev_mc_list	*node = adapter->net_dev->mc_list;
	u8			counter;

	if(!TEST_FLAG(adapter, ACCEPT_MULTICAST))
			return -EINVAL;
	if(TEST_FLAG(adapter, ACCEPT_ALLMULTICAST))
			return 0;

	/*
	 * We accept Multicast, but we do not accept them all.
	 * We need to check the filter to check if we accept this specific address.
	 */
	for(counter = 0; counter < adapter->net_dev->mc_count; counter++){
		if(node && memcmp(node->dmi_addr, addr, ETH_ALEN) == 0)
			return 0;
		node = node->next;
	}

	/*
	 * We haven't found the address, we must drop this packet.
	 */
	return -EINVAL;
}

static inline char rt2x00_filter_unicast(struct _rt2x00_adapter *adapter, char *addr)
{
	if(!TEST_FLAG(adapter, ACCEPT_UNICAST))
		return -EINVAL;

	if(TEST_FLAG(adapter, ACCEPT_PROMISC))
		return 0;

	if(memcmp(&adapter->net_dev->dev_addr, addr, ETH_ALEN) != 0)
		return -EINVAL;

	return 0;
}

char rt2x00_filter_packet(struct _rt2x00_adapter *adapter, struct sk_buff *skb)
{
	struct ieee802_11_hdr	*header802_11 = (struct ieee802_11_hdr*)skb->data;
	
	struct _skb_cb_rx	*skb_cb = (struct _skb_cb_rx*)skb->cb;

	/*
	 * Per mode packet requirements:
	 * AD-HOC - 
	 *  - Addr3 should be our address, unless promisc mode is enabled.
	 *  - Packet should not come from or send to AP.
	 *  - If multicast, addr2 should be a allowed source.
	 *
	 * Infrastructure - 
	 *  - Addr1 should be our address, unless promisc mode is enabled.
	 *  - Packet should come from AP.
	 *  - If multicast, addr2 should be a allowed source.
	 *
	 * Master - 
	 *  - Addr2 must be our address.
	 *  - Packet must be send to AP.
	 *  - If multicast, addr3 should be a allowed source.
	 */ 

	if(LINK_TYPE(adapter, IW_MODE_ADHOC)){
		if(skb_cb->rxd.unicast && !rt2x00_filter_unicast(adapter, header802_11->addr3))
			return -EINVAL;
		if(skb_cb->rxd.multicast || skb_cb->rxd.broadcast)
			if(!rt2x00_filter_multicast(adapter, header802_11->addr2))
				return -EINVAL;
		if(header802_11->frame_ctl & IEEE802_11_FCTL_FROMDS || header802_11->frame_ctl & IEEE802_11_FCTL_TODS)
			return -EINVAL;
    	} else if(LINK_TYPE(adapter, IW_MODE_INFRA)){
		/* if this is a beacon then we probably want to know about it*/
		if(header802_11->frame_ctl == IEEE802_11_STYPE_BEACON)
		    return 0;
		if(skb_cb->rxd.unicast && !rt2x00_filter_unicast(adapter, header802_11->addr1))
			return -EINVAL;
		if(skb_cb->rxd.multicast || skb_cb->rxd.broadcast)
			if(!rt2x00_filter_multicast(adapter, header802_11->addr2))
				return -EINVAL;
		if(!(header802_11->frame_ctl & IEEE802_11_FCTL_FROMDS))
			return -EINVAL;
	} else if(LINK_TYPE(adapter, IW_MODE_MASTER)){
		if(skb_cb->rxd.unicast && !rt2x00_filter_unicast(adapter, header802_11->addr2))
			return -EINVAL;
		if(skb_cb->rxd.multicast || skb_cb->rxd.broadcast)
			if(!rt2x00_filter_multicast(adapter, header802_11->addr3))
				return -EINVAL;
		if(!(header802_11->frame_ctl & IEEE802_11_FCTL_TODS))
			return -EINVAL;
	}

	/*
	 * If packet is encrypted we must have encryption enabled to be able to accept packet.
	 * Unencrypted packets cannot be accepted when encryption type is set to restricted. 
	 */
	if(header802_11->frame_ctl & IEEE802_11_FCTL_WEP && !TEST_FLAG(adapter, CRYPTO_ENABLED))
		return -EINVAL;

	if(!(header802_11->frame_ctl & IEEE802_11_FCTL_WEP) && TEST_FLAG(adapter, CRYPTO_RESTRICT))
		return -EINVAL;

	if(header802_11->frame_ctl & IEEE802_11_STYPE_NULLFUNC ||
	   header802_11->frame_ctl & IEEE802_11_STYPE_DATA_CFACK ||
	   header802_11->frame_ctl & IEEE802_11_STYPE_CFPOLL ||
	   header802_11->frame_ctl & IEEE802_11_STYPE_CFACKPOLL)
		return -EINVAL;


	return 0;
}

static inline char rt2x00_repeat_packet(struct _rt2x00_adapter *adapter, struct sk_buff *skb)
{
	struct _skb_cb_tx	*skb_cb = (struct _skb_cb_tx*)skb->cb;

	/*
	 * Flag packet as repeater_packet to make sure this packet is send without
	 * any changes during tx processing.
	 */
	 
	skb_cb->repeater_packet = 1;

	skb = rt2x00_update_802_11hdr(adapter, skb);

	if(skb==NULL){   
	    return -1; /* memory problem, so abort */
	}
	
	spin_lock_bh(&adapter->tx_lock);

	skb_queue_tail(&adapter->tx_packet_queue, skb);
	
	spin_unlock_bh(&adapter->tx_lock);

	return 0;
}

static inline char rt2x00_merge_packet(struct _rt2x00_adapter *adapter, struct sk_buff *main, struct sk_buff *frag)
{
	struct sk_buff		*tmp;
	struct _skb_cb_rx	*skb_cb = (struct _skb_cb_rx*)frag->cb;
	u16			frag_length = frag->len - skb_cb->header_length;

	/*
	 * Remove both fragments from their lists.
	 */
	skb_unlink(frag);
	skb_unlink(main);

	/*
	 * Check if we are not going to exceed the fragment size.
	 */
	if((main->len + frag_length) > MAX_PACKET_SIZE){
		
		tmp = main;
		main = skb_copy_expand(main, 0, frag_length, GFP_ATOMIC);
		dev_kfree_skb_irq(tmp);
		if(!main)
			return -ENOMEM;
	}

	memcpy(skb_put(main, frag_length), frag->data + skb_cb->header_length, frag_length);

	/*
	 * Free the no longer needed fragment.
	 */
	dev_kfree_skb_irq(frag);

	/*
	 * Place the main fragment back to the list.
	 */
	skb_queue_tail(&adapter->frag_packet_queue, main);

	return 0;
}

struct sk_buff* rt2x00_construct_packet_fragments(struct _rt2x00_adapter *adapter, struct sk_buff *skb)
{
	struct sk_buff		*main_packet;
	struct ieee802_11_hdr	*header802_11 = (struct ieee802_11_hdr*)skb->data;
	struct _skb_cb_rx	*skb_cb = (struct _skb_cb_rx*)skb->cb;

	/*
	 * Is this the first  of multiple fragments.
	 */
	if((header802_11->seq_ctl & IEEE802_11_SCTL_FRAG) == 0){
		skb_cb->seq_ctl = header802_11->seq_ctl;
		skb_queue_tail(&adapter->frag_packet_queue, skb);
		return skb;
	}

	/*
	 * We have to search the queue to find the next fragment.
	 */
	skb_queue_walk(&adapter->frag_packet_queue, main_packet){
		skb_cb = (struct _skb_cb_rx*)main_packet->cb;

		if((header802_11->seq_ctl - skb_cb->seq_ctl) == 1){
			if(rt2x00_merge_packet(adapter, main_packet, skb))
				return NULL;
			else
				goto exit;
		} else if(((skb_cb->seq_ctl - header802_11->seq_ctl) & IEEE802_11_SCTL_SEQ) == 0){
			/*
			 * We missed one of the fragments. And we have made sure this is the same sequence.
			 * We cannot complete this packet so we must drop it.
			 */
			skb_unlink(main_packet);
			dev_kfree_skb_irq(main_packet);
			goto fail;
		}
	}

fail:
	adapter->stats.rx_missed_errors++;
	return NULL;

exit:
	return main_packet;
}

char rt2x00_receive_packet(struct _rt2x00_adapter *adapter, struct sk_buff *skb)
{
	struct ieee802_11_hdr	*header802_11 = (struct ieee802_11_hdr*)skb->data;

	/* It is necessary to set up all the skb parameters even
	   in monitor mode */
	   
	skb->mac.raw = skb->data;
	skb->pkt_type = PACKET_OTHERHOST;
	skb->ip_summed = CHECKSUM_NONE;
	skb->dev = adapter->net_dev;

	/* Don't filter packet, accept all, and do not change header */
	if(LINK_TYPE(adapter, IW_MODE_MONITOR))
		return 0;

	if(rt2x00_filter_validate(adapter, skb))
		return -EINVAL;

	if(rt2x00_filter_duplicate(adapter, skb))
		return -EINVAL;

	/* Validate packet, update header, retransmit packet. */
	if(LINK_TYPE(adapter, IW_MODE_REPEAT))
		return rt2x00_repeat_packet(adapter, skb);

	if(rt2x00_filter_packet(adapter, skb))
		return -EINVAL;

	if((header802_11->frame_ctl & IEEE802_11_FCTL_WEP)) 
	    if(rt2x00_decrypt(adapter, skb))
		return -EINVAL;

	/*
	 * Re-construct fragmented packets.
	 * If the MOREFRAGS flags is not set, but the fragment number is higher then 0
	 * it means we have reached the end of a series.
	 */
	if(header802_11->frame_ctl & IEEE802_11_FCTL_MOREFRAGS){
		if(rt2x00_construct_packet_fragments(adapter, skb))
			return -EINVAL;
		/* We expect more fragments so do not go any further. */
		return 0;
	} else if((header802_11->seq_ctl & IEEE802_11_SCTL_FRAG) > 0){
		skb = rt2x00_construct_packet_fragments(adapter, skb);
		if(!skb)
			return -EINVAL;
		/*
		 * We have successfully acquired the complete packet, unlink it from the fragment queue.
		 */
		skb_unlink(skb);
	}

	if(!(header802_11->frame_ctl & IEEE802_11_FTYPE_DATA) && !(header802_11->frame_ctl & IEEE802_11_FTYPE_CTL))
		return rt2x00_receive_mgmt_packet(adapter, skb);


	/* Replace 802.11  header with the 802.3 header. */
	rt2x00_construct_802_3hdr(adapter, skb);

	return 0;
}

static inline void rt2x00_update_tx_stats(struct _rt2x00_adapter *adapter, struct _txd *txd)
{
	switch(txd->tx_result){
		case TX_SUCCESS_RETRY:
			adapter->stats.collisions += txd->retry_count;
			/* Fall through because TX_SUCCESS_RETRY implies TX_SUCESS.*/
		case TX_SUCCESS:
			adapter->stats.tx_packets++;
			if(adapter->chip_adapter == RT2400)
				adapter->stats.tx_bytes += txd->chip.rt2400.data_byte_count;
			else
				adapter->stats.tx_bytes += txd->data_byte_count;
		break;
		case TX_FAIL_RETRY:
			adapter->iw_stats.discard.retries++;
			adapter->stats.tx_errors++;
			adapter->stats.collisions += txd->retry_count;
		break;	
		case TX_FAIL_INVALID:
		case TX_FAIL_OTHER:
			adapter->iw_stats.discard.misc++;
			adapter->stats.tx_errors++;
		break;
	}
}

void rt2x00_interrupt_tcbn_expire(struct _rt2x00_adapter *adapter)
{
	struct ieee802_11_mgmt_hdr	header802_11;

	spin_lock(&adapter->beacon_lock);
	device_to_cpu(&header802_11, adapter->dma->beacon_ring.desc_addr, sizeof(header802_11));
	FRAME_SEQ_INC(adapter);					/* update BEACON frame's sequence number */
	header802_11.seq_ctl = adapter->config->sequence;	/* Sequence number. */
	cpu_to_device(adapter->dma->beacon_ring.desc_addr, &header802_11, sizeof(header802_11));
	spin_unlock(&adapter->beacon_lock);
}

void rt2x00_interrupt_twake_expire(struct _rt2x00_adapter *adapter)
{








}

void rt2x00_interrupt_tatimw_expire(struct _rt2x00_adapter *adapter)
{









}

void rt2x00_interrupt_txdone_txring(struct _rt2x00_adapter *adapter)
{
	struct _txd		txd;

	spin_lock(&adapter->tx_lock);

	while(adapter->dma->tx_ring_done_index != adapter->dma->tx_ring_index){
		device_to_cpu(&txd, adapter->dma->tx_ring[adapter->dma->tx_ring_done_index].desc_addr, sizeof(struct _txd));

		if(txd.owner == DESC_OWNED_BY_NIC || !txd.valid)
			break;

		rt2x00_update_tx_stats(adapter, &txd);

		txd.valid = 0;
		cpu_to_device(adapter->dma->tx_ring[adapter->dma->tx_ring_done_index].desc_addr, &txd, sizeof(struct _txd));

		INCREASE_RING(adapter->dma->tx_ring_done_index, TX_RING_SIZE);
	}

	if(!skb_queue_empty(&adapter->tx_packet_queue))
		tasklet_schedule(&adapter->tx_tasklet);

	spin_unlock(&adapter->tx_lock);
}

void rt2x00_interrupt_txdone_atimring(struct _rt2x00_adapter *adapter)
{






}

void rt2x00_interrupt_txdone_prioring(struct _rt2x00_adapter *adapter)
{
	struct _txd		txd;

	spin_lock(&adapter->prio_lock);

	while(adapter->dma->prio_ring_done_index != adapter->dma->prio_ring_index){
		device_to_cpu(&txd, adapter->dma->prio_ring[adapter->dma->prio_ring_done_index].desc_addr, sizeof(struct _txd));

		if(txd.owner == DESC_OWNED_BY_NIC || !txd.valid)
			break;

		rt2x00_update_tx_stats(adapter, &txd);

		txd.valid = 0;
		cpu_to_device(adapter->dma->prio_ring[adapter->dma->prio_ring_done_index].desc_addr, &txd, sizeof(struct _txd));

		INCREASE_RING(adapter->dma->prio_ring_done_index, PRIO_RING_SIZE);
	}

	if(!skb_queue_empty(&adapter->prio_packet_queue))
		tasklet_schedule(&adapter->prio_tasklet);

	spin_unlock(&adapter->prio_lock);
}

void rt2x00_interrupt_rxdone(struct _rt2x00_adapter *adapter)
{
	struct _rxd		rxd;
	u8			*dma = NULL;
	struct ieee802_11_hdr	*header802_11 = NULL;
	struct sk_buff		*skb = NULL;
	struct _skb_cb_rx	*skb_cb = NULL;

	spin_lock(&adapter->rx_lock);

	while(adapter->dma->rx_ring_done_index != adapter->dma->rx_ring_index){
		dma = (u8*) adapter->dma->rx_ring[adapter->dma->rx_ring_index].data_addr;
		device_to_cpu(&rxd, adapter->dma->rx_ring[adapter->dma->rx_ring_index].desc_addr, sizeof(struct _rxd));
		header802_11 = (struct ieee802_11_hdr*)dma; 

		if(rxd.owner == DESC_OWNED_BY_NIC)
			break;

		/* If it is the first frame of multiple fragments allocate the maximum frame_size. */
		if(header802_11->frame_ctl & IEEE802_11_FCTL_MOREFRAGS &&
		  !(header802_11->seq_ctl & IEEE802_11_SCTL_FRAG))
			skb = dev_alloc_skb(MAX_PACKET_SIZE);
		else
			skb = dev_alloc_skb(rxd.data_byte_count + 2);
		if(!skb){
			adapter->stats.rx_dropped++;
			/* Hand dma back to the adapter. */
			rxd.owner = DESC_OWNED_BY_NIC;
			cpu_to_device(adapter->dma->rx_ring[adapter->dma->rx_ring_index].desc_addr, &rxd, sizeof(struct _rxd));
			break;
		}

		skb_reserve(skb, 2);	/* 16 byte align the IP header */
						
		skb_cb = (struct _skb_cb_rx*)skb->cb;
		memset(skb_cb, 0x00, sizeof(struct _skb_cb_rx));

		memcpy(&skb_cb->rxd, &rxd, sizeof(struct _rxd));
		skb_cb->header_length = IEEE802_11_HLEN;		
		
		/* Copy entire frame to buffer. */
		device_to_cpu(skb_put(skb, rxd.data_byte_count), dma, rxd.data_byte_count);

		/* Enqueue packet for processing. */
		skb_queue_tail(&adapter->rx_packet_queue, skb);

		/* Hand dma back to the adapter. */
		rxd.owner = DESC_OWNED_BY_NIC;
		cpu_to_device(adapter->dma->rx_ring[adapter->dma->rx_ring_index].desc_addr, &rxd, sizeof(struct _rxd));

		/* Move to next packet in dma-> */
		INCREASE_RING(adapter->dma->rx_ring_index, RX_RING_SIZE);
	}

	/* 
	 * Update the done index. 
	 * Make sure the rx done index is 1 lower then main index.
	 * Also make sure to check for end of buffer.
	 */
	if(adapter->dma->rx_ring_index == 0)
		adapter->dma->rx_ring_done_index = RX_RING_SIZE;
	else
		adapter->dma->rx_ring_done_index = adapter->dma->rx_ring_index - 1;

	/* Schedule rx packet processing. */
	if(!skb_queue_empty(&adapter->rx_packet_queue))
		tasklet_schedule(&adapter->rx_tasklet);

	spin_unlock(&adapter->rx_lock);
}

void rt2x00_write_tx_descriptor(struct _rt2x00_adapter *adapter, struct _txd *txd, u8 *desc, u32 rate, u32 length)
{
	struct _txd	tmp;
	u16		length_high = 0;
	u16		length_low = 0;
	u32		residual = 0;

	txd->owner = DESC_OWNED_BY_NIC;
	txd->valid = 1;
	txd->retry_mode = adapter->config->long_retry;
	txd->timestamp = 0;
	if(adapter->chip_adapter == RT2500)
		txd->rt2500.tx_rate = rate;

	/*
	 * The descriptors contain several BBP register values.
	 * When setting the value pay attention to the notation:
	 * 0x8000: Set the BBP register entry to valid.
	 * x << 8: Set BBP register number to the correct offset. See BBPCSR for correct bits.
	 * After that you can set the value that you want to write to this BBP register.
	 */
	if(adapter->chip_adapter == RT2400){
		txd->chip.rt2400.data_byte_count = length;
		txd->chip.rt2400.plcp_signal = 0x8000 | 5 << 8 | _BITPLCP(adapter, rate);
		if(adapter->config->preamble == PREAMBLE_SHORT)
			txd->chip.rt2400.plcp_signal |= 0x0008; /* Enable short preamble. */
	
		txd->chip.rt2400.plcp_service = 0x8000 | 6 << 8 | 0x0004;

		txd->chip.rt2400.bbcr4 = 0x8000 | 3 << 8 | adapter->config->tx_power;
	} else {
		txd->data_byte_count = length;
		txd->chip.rt2500.plcp_signal = 0x8000 | 5 << 8 | _BITPLCP(adapter, rate);
		if(adapter->config->preamble == PREAMBLE_SHORT)
			txd->chip.rt2500.plcp_signal |= 0x0008; /* Enable short preamble. */
	
		/*
		 * Working with tx_powers is not well documented in the RT2400 drivers.
		 * That it should be set in the register here is just guess.
		 */
		txd->chip.rt2500.plcp_service = 0x04;

		/* For RT2500 devices we have to notify the device if we are working with a OFDM rate. */
		txd->rts_ofdm = (rate < 4) ?  1 : 0;

		/*
		 * We do not work with priority packets like the rt2500 driver.
		 * So we just take the default values here.
		 */
		txd->chip.rt2500.cw_min = 5;	/* 2^5 = 32 */
		txd->chip.rt2500.cw_max = 10;	/* 2^10 = 1024 */
		txd->chip.rt2500.aifs = 1;

		/*
		 * RT2500 devices have a second interrupt for encryption and decryption.
		 * We are not going to use it, so the owner will always be HOST to inform
		 * the adapter we are not using his encryption interrupts.
		 */
		txd->cipher_owner = DESC_OWNED_BY_HOST;
	}

	length += 4;

	if(adapter->config->cur_rate < 4){ /* OFDM Rate. */
		/* Convert the length to microseconds. */
		length = DURATION(length, _BITRATE(adapter, rate));
		residual = DURATION_RES(length, _BITRATE(adapter, rate));
		if(residual != 0)
			length++;

		length_high = 0x8000 | 7 << 8 | (length >> 8);
		length_low = 0x8000 | 8 << 8 | (length & 0xff);
	} else {
		length_high = (length >> 4);
		length_low = (length & 0xf);
	}

	if(adapter->chip_adapter == RT2400){
		txd->chip.rt2400.plcp_length_high = length_high;
		txd->chip.rt2400.plcp_length_low = length_low;
		/* 802.11b standard 18.2.3.5. Check for need to set length extension. */
		if(residual <= (8 % 11))
			txd->chip.rt2400.plcp_signal |= 0x0080;
	} else {
		txd->chip.rt2500.plcp_length_high = length_high;
		txd->chip.rt2500.plcp_length_low = length_low;
		/* 802.11b standard 18.2.3.5. Check for need to set length extension. */
		if(residual <= (8 % 11))
			txd->chip.rt2500.plcp_signal |= 0x0080;
	}

	/*
	 * Convert descriptor to 32bit litle endian.
	 * Copy the data to the descriptor, but be aware that the last byte of the descriptor
	 * should be written seperatly in case of RT2500 devices.
	 * This because the last byte should be written with a larger offset.
	 * For RT2400 devices it will not be needed since that byte will be unused anyway.
	 */
	cpu_to_device(&tmp, txd, sizeof(struct _txd));
	memcpy(desc, &tmp, sizeof(struct _txd) - sizeof(u32));
	if(adapter->chip_adapter == RT2500)
		memcpy((desc + sizeof(struct _txd) - sizeof(u32) + TXD_SPACE), &tmp.rt2500, sizeof(u32));
}

/*
 * rt2x00_test_free_descriptors()
 *  Check if there are enough free descriptors ready for use. 
 *  A descriptor is free and ready when the valid bit is not set,
 *  and when the owner bit is set to DESC_OWNED_BY_HOST.
 *  Since the descriptors are filled and send in chronological order,
 *  we only have to test the first and last descriptor.
 */
static inline u8 rt2x00_test_descriptors(struct _dma_ring (*desc_ring)[], u8 desc_index, u8 desc_max_size, u8 fragments)
{
	struct _txd	txd;

	device_to_cpu(&txd, ((*desc_ring)[desc_index]).desc_addr, sizeof(struct _txd));
	if((txd.owner != DESC_OWNED_BY_HOST) || txd.valid)
		return -ENOMEM;

	desc_index += fragments - 1; /* Move to the next descriptor. */
	if(desc_index > desc_max_size)
		desc_index -= desc_max_size;

	device_to_cpu(&txd, ((*desc_ring)[desc_index]).desc_addr, sizeof(struct _txd));
	if((txd.owner != DESC_OWNED_BY_HOST) || txd.valid)
		return -ENOMEM;

	return 0;
}

u8 rt2x00_test_free_descriptors(struct _rt2x00_adapter *adapter, u8 ring, u8 fragments)
{
	u8 status = 0;

	if(fragments == TX_RING){
		spin_lock(&adapter->tx_lock);
		status = rt2x00_test_descriptors(&adapter->dma->tx_ring, adapter->dma->tx_ring_index, TX_RING_SIZE, fragments);
		spin_unlock(&adapter->tx_lock);
	} else if(fragments == PRIO_RING){
		spin_lock(&adapter->prio_lock);
		status = rt2x00_test_descriptors(&adapter->dma->prio_ring, adapter->dma->prio_ring_index, PRIO_RING_SIZE, fragments);
		spin_unlock(&adapter->prio_lock);
	}

	return status;
}

char rt2x00_create_rts_packet(struct _rt2x00_adapter *adapter, struct sk_buff *skb)
{
	struct _txd		txd;
	u8			*dma = (u8*) adapter->dma->tx_ring[adapter->dma->tx_ring_index].data_addr;
	struct _skb_cb_tx	*skb_cb = (struct _skb_cb_tx*)skb->cb;
	struct ieee802_11_hdr	*header802_11 = (struct ieee802_11_hdr*)skb->data;
	u32			size;

	device_to_cpu(&txd, adapter->dma->tx_ring[adapter->dma->tx_ring_index].desc_addr, sizeof(struct _txd));

	if(txd.owner == DESC_OWNED_BY_NIC || txd.valid)
		return -ENOMEM;

	header802_11->frame_ctl |= IEEE802_11_FTYPE_CTL | IEEE802_11_STYPE_RTS;

	if(skb_cb->fragments > 1)
		size = adapter->config->fragmentation;
	else
		size = skb->len;

	/* 
	 * Calculate duration.
	 * 2 SIFS + header + frame size 
	 */
	header802_11->duration_id = (2 * adapter->config->sifs)
				+ rt2x00_packet_duration(adapter, BASIC_RATE(adapter), IEEE802_11_HLEN);
				+ rt2x00_packet_duration(adapter, BITRATE(adapter), size - IEEE802_11_HLEN);

	device_to_cpu(dma, header802_11, sizeof(struct ieee802_11_hdr));

	if(adapter->chip_adapter == RT2400)
		txd.rts_ofdm = 1;
	else
		txd.rt2500.rts = 1;

	txd.ack = 1;
	txd.more_frag = 1;
	txd.ifs = IFS_BACKOFF;

	rt2x00_write_tx_descriptor(
		adapter, &txd, (u8*)adapter->dma->tx_ring[adapter->dma->tx_ring_index].desc_addr, BASIC_RATE(adapter), size);

	INCREASE_RING(adapter->dma->tx_ring_index, TX_RING_SIZE);

	return 0;
}

char rt2x00_create_repeater_packet(struct _rt2x00_adapter *adapter, struct sk_buff *skb, u8 *desc, u8 *dma)
{
	struct _txd	txd;
	device_to_cpu(&txd, desc, sizeof(struct _txd));

	if(txd.owner == DESC_OWNED_BY_NIC || txd.valid)
		return -ENOMEM;

	cpu_to_device(dma, &skb->data, skb->len);

	txd.ack = 0;
	txd.more_frag = 0;
	txd.ifs = IFS_SIFS;

	rt2x00_write_tx_descriptor(adapter, &txd, desc, BITRATE(adapter), skb->len);

	INCREASE_RING(adapter->dma->tx_ring_index, TX_RING_SIZE);

	return 0;
}

char rt2x00_create_packet(struct _rt2x00_adapter *adapter, struct sk_buff *skb, u8 framegap)
{
	struct _txd		txd;
	u8			*dma = (u8*) adapter->dma->tx_ring[adapter->dma->tx_ring_index].data_addr;
	u8			*desc = (u8*) adapter->dma->tx_ring[adapter->dma->tx_ring_index].desc_addr;
	struct _skb_cb_tx	*skb_cb = (struct _skb_cb_tx*)skb->cb;
	struct ieee802_11_hdr	*header802_11 = (struct ieee802_11_hdr*)skb->data;
	u32			size = skb->len - skb_cb->header_length;
	u16			offset;

	device_to_cpu(&txd, desc, sizeof(struct _txd));

	if(txd.owner == DESC_OWNED_BY_NIC || txd.valid)
		return -ENOMEM;

	if(skb_cb->multicast){
		/* No ACK expected for multicast frame */
		header802_11->duration_id = 0;
		txd.ack = 0;
	} else if(skb_cb->fragments > 2){
		header802_11->frame_ctl |= IEEE802_11_FCTL_MOREFRAGS;
		header802_11->duration_id = (2 * adapter->config->sifs)
					+ rt2x00_packet_duration(adapter, BASIC_RATE(adapter), skb_cb->header_length)
					+ rt2x00_packet_duration(adapter, BITRATE(adapter), adapter->config->fragmentation);
	} else if(skb_cb->fragments == 2){
		header802_11->duration_id = (3 * adapter->config->sifs)
					+ rt2x00_packet_duration(adapter, BASIC_RATE(adapter), skb_cb->header_length)
					+ rt2x00_packet_duration(adapter, BITRATE(adapter), size);
	} else {
		header802_11->duration_id = adapter->config->sifs
					+ rt2x00_packet_duration(adapter, BASIC_RATE(adapter), skb_cb->header_length);
	}

	cpu_to_device(dma, header802_11, sizeof(struct ieee802_11_hdr));

	/*
	 * Calculate which part of the packet we are going to send now.
	 */
	offset = (header802_11->seq_ctl & IEEE802_11_SCTL_FRAG) * adapter->config->fragmentation;

	/* We will start writing directly after the header. */
	dma += skb_cb->header_length;

	/* When encryption is enabled we must always copy the first 4 bytes a first. */
	if(header802_11->frame_ctl & IEEE802_11_FCTL_WEP){
		cpu_to_device(dma, (skb->data + skb_cb->header_length), 4);
		dma += 4;
		offset += 4;
	}

	if(skb_cb->fragments == 1)
		cpu_to_device(dma, (skb->data + offset), (skb->len - offset));
	else
		cpu_to_device(dma, (skb->data + offset), adapter->config->fragmentation);

	/* When encryption is enabled we must always copy the last 4 bytes at the end of the packet too. */
	if(header802_11->frame_ctl & IEEE802_11_FCTL_WEP)
		cpu_to_device(dma, (skb->data + (skb->len - 4)), 4);

	txd.more_frag = (skb_cb->fragments > 1);
	txd.ifs = framegap;

	rt2x00_write_tx_descriptor(adapter, &txd, desc, BITRATE(adapter), size);

	INCREASE_RING(adapter->dma->tx_ring_index, TX_RING_SIZE);

	return 0;
}

char rt2x00_start_transmit(struct _rt2x00_adapter *adapter, struct sk_buff *skb)
{
	struct ieee802_11_hdr	*header802_11 = (struct ieee802_11_hdr*)skb->data;
	struct _skb_cb_tx	*skb_cb = (struct _skb_cb_tx*)skb->cb;
	u8			framegap = IFS_BACKOFF;

	spin_lock_bh(&adapter->tx_lock);
	read_lock(&adapter->config_lock);

	if(rt2x00_test_free_descriptors(adapter, TX_RING, skb_cb->fragments + skb_cb->rts)){
		read_unlock(&adapter->config_lock);
		spin_unlock_bh(&adapter->tx_lock);
		return -ENOMEM;
	}

	if(skb_cb->repeater_packet){
		if(rt2x00_create_repeater_packet(adapter, skb,
		   adapter->dma->tx_ring[adapter->dma->tx_ring_index].desc_addr,
		   adapter->dma->tx_ring[adapter->dma->tx_ring_index].data_addr)){
			read_unlock(&adapter->config_lock);
			spin_unlock_bh(&adapter->tx_lock);
			return -ENOMEM;
		}
		goto exit;
	}

	FRAME_SEQ_INC(adapter);

	header802_11->seq_ctl = 0;				/* Fragment number. */
	header802_11->seq_ctl |= adapter->config->sequence;	/* Sequence number. */


	/* If multiple fragment required, RTS is required only for the first fragment. */
	if(skb_cb->rts){
		if(rt2x00_create_rts_packet(adapter, skb)){
			read_unlock(&adapter->config_lock);
			spin_unlock_bh(&adapter->tx_lock);
			return -ENOMEM;
		}
		framegap = IFS_SIFS;	/* Init frame gap for coming data after RTS. */
	}

	/*
	 * Start Copy Ndis Packet into Ring buffer.
	 * For frame required more than one ring buffer (fragment), all ring buffers
	 * have to be filled before kicking start tx bit.
	 */
	while(skb_cb->fragments){
		if(rt2x00_create_packet(adapter, skb, framegap)){
			read_unlock(&adapter->config_lock);
			spin_unlock_bh(&adapter->tx_lock);
			return -ENOMEM;
		}
		framegap = IFS_SIFS;
		--skb_cb->fragments;
		header802_11->seq_ctl = (header802_11->seq_ctl + 0x0001) & IEEE802_11_SCTL_FRAG; /* Increase fragment number. */
	}

	/* Acknowledge protocol send complete of pending packet. */
	dev_kfree_skb_irq(skb);

exit:
	read_unlock(&adapter->config_lock);
	spin_unlock_bh(&adapter->tx_lock);

	return 0;
}

char rt2x00_enqueue_packet(struct _rt2x00_adapter *adapter, struct sk_buff *skb)
{
	struct ieee802_11_hdr	*header802_11;
	struct _skb_cb_tx	*skb_cb = (struct _skb_cb_tx*)skb->cb;

	read_lock(&adapter->config_lock);

	memset(skb->cb, 0x00, sizeof(skb->cb));

	/* Replace the 802.3 header with the 802.11 header. */
	skb = rt2x00_construct_802_11hdr(adapter, skb);
	if(!skb){
		read_unlock(&adapter->config_lock);
		return -EINVAL;
	}

	/* Update pointers. */
	header802_11 = (struct ieee802_11_hdr*)skb->data;
	skb_cb = (struct _skb_cb_tx*)skb->cb;

	if(ADDR_IS_MULTICAST(header802_11->addr1)){ 
		skb_cb->multicast = 1; /* Packet is broadcast or multicast. */
		skb_cb->fragments = 1; /* If multicast only 1 fragments will be send. */
	} else {
		/* Calculate fragments required. */
		skb_cb->fragments = (skb->len) / (adapter->config->fragmentation);
		/* If the size does not exactly fit into the allowable fragments size add 1 fragment. */
		if((skb->len) % (adapter->config->fragmentation))
			skb_cb->fragments++;
	}

	/* If the fragment size large than RTS threshold, RTS is required. */
	if (skb_cb->fragments > 1)
		skb_cb->rts = (adapter->config->fragmentation > adapter->config->rts_treshold);
	else
		skb_cb->rts = (skb->len > adapter->config->rts_treshold);

	if(header802_11->frame_ctl & IEEE802_11_FCTL_WEP)
		rt2x00_encrypt(adapter, skb); /* Encrypt the frame. */

	read_unlock(&adapter->config_lock);
	spin_lock_bh(&adapter->tx_lock);

	/* Enqueue the packet to our tx queue, so it can be transmitted when the adapter is ready. */
	skb_queue_tail(&adapter->tx_packet_queue, skb);

	spin_unlock_bh(&adapter->tx_lock);
	return 0;
}

void rt2x00_delayed_start_tx(unsigned long data)
{
	struct _rt2x00_adapter		*adapter = (struct _rt2x00_adapter*) data;
	struct sk_buff			*skb = NULL;

	spin_lock_bh(&adapter->tx_lock);

	while(!skb_queue_empty(&adapter->tx_packet_queue)){

		if(!TEST_FLAG(adapter, ADAPTER_RADIO_ON) || !TEST_FLAG(adapter, ADAPTER_ENABLED))
			break; /* Radio off or we are not ready, abort sending packets. */

		skb = skb_dequeue(&adapter->tx_packet_queue);

		if(rt2x00_start_transmit(adapter, skb)){
			DEBUG("Not enough descriptors available.\n");
			skb_queue_head(&adapter->tx_packet_queue, skb);
			break;
		}
	}

	spin_unlock_bh(&adapter->tx_lock);

	write_lock(&adapter->config_lock);
	/* Kick TX Control Register at the end of all ring buffer preparation */
	rt2x00_register_write(adapter, TXCSR0, 0x0001);
	write_unlock(&adapter->config_lock);
}

void rt2x00_delayed_start_tx_prio(unsigned long data)
{
	struct _rt2x00_adapter		*adapter = (struct _rt2x00_adapter*) data;
	struct sk_buff			*skb = NULL;

	spin_lock_bh(&adapter->prio_lock);

	while(!skb_queue_empty(&adapter->prio_packet_queue)){

		if(!TEST_FLAG(adapter, ADAPTER_RADIO_ON) || !TEST_FLAG(adapter, ADAPTER_ENABLED))
			break; /* Radio off or we are not ready, abort sending packets. */

		skb = skb_dequeue(&adapter->prio_packet_queue);

		if(rt2x00_transmit_mgmt(adapter, skb)){
			DEBUG("Not enough descriptors available.\n");
			skb_queue_head(&adapter->prio_packet_queue, skb);
			break;
		}
	}

	spin_unlock_bh(&adapter->prio_lock);

	write_lock(&adapter->config_lock);
	/* Kick PRIO Control Register at the end of all ring buffer preparation */
	rt2x00_register_write(adapter, TXCSR0, 0x0004);
	write_unlock(&adapter->config_lock);
}

void rt2x00_delayed_handle_rx(unsigned long data)
{
	struct _rt2x00_adapter		*adapter = (struct _rt2x00_adapter*) data;
    	struct sk_buff			*skb = NULL;

	spin_lock_bh(&adapter->rx_lock);

	while(!skb_queue_empty(&adapter->rx_packet_queue)){

		if(!TEST_FLAG(adapter, ADAPTER_ENABLED))
			break; /* We are not ready, abort sending packets. */

		skb = skb_dequeue(&adapter->rx_packet_queue);
    
		adapter->stats.rx_packets++;
		adapter->stats.rx_bytes += skb->len;
		if(((struct _skb_cb_rx*)skb)->multicast)
			adapter->stats.multicast++;
	
		if(rt2x00_receive_packet(adapter, skb)==0){		/* Packet received succesfull. */
			if(!LINK_TYPE(adapter,IW_MODE_REPEAT))		/* We must not send a repeat skb to the net_rx or we get a double kfree */
			    netif_rx(skb);				/* Send packet to upper layer. */
			adapter->net_dev->last_rx = jiffies;		/* Update RX time. */
		} else {						/* Packet failed. */
			adapter->stats.rx_errors++;			/* Update error count. */
			dev_kfree_skb_irq(skb);				/* Release skb packet. */
		}
	}

	spin_unlock_bh(&adapter->rx_lock);

	if(LINK_TYPE(adapter, IW_MODE_REPEAT) && !skb_queue_empty(&adapter->tx_packet_queue))
		tasklet_schedule(&adapter->tx_tasklet);
}
