/*
	rt2x00 SourceForge Project - <http://rt2x00.serialmonkey.com>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the
	Free Software Foundation, Inc.,
	59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

	Licensed under the GNU GPL
	Original code supplied under license from RaLink Inc, 2003
 */

/*
	Module Name: rt2x00_mgmt.c

	Abstract: Management packet handlers.

	Revision History:
	Who		When		What
	--------	-----------	-----------------------------
	IvD		<releasedate>	Initial Release.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/timer.h>

#include <net/iw_handler.h>
#include <ieee802_11.h>

#include "rt2x00_mgmt.h"
#include "rt2x00_adapter.h"

void rt2x00_scan_update(unsigned long data)
{
	struct _rt2x00_adapter	*adapter = (struct _rt2x00_adapter*)data;
	char			update = 1;

	/*
	 * Check if all frequencies has to be scanned or not.
	 */
	if(!TEST_FLAG(&adapter->scan, IW_SCAN_ALL_FREQ)){
		rt2x00_end_scan(adapter);
		return;
	}

	write_lock(&adapter->config_lock);

	/*
	 * Check to see if we have reached the last channel.
	 */
	if(adapter->config->channel < adapter->capability->size_channels)
		adapter->config->channel++;
	else 
		update = 0;
	

	write_unlock(&adapter->config_lock);

	if(update){
		INFO("Scanning all channels, switching to next channel %d of %d...\n", adapter->config->channel, adapter->capability->size_channels);
		/*
		 * Update adapter with new channel.
		 */
		rt2x00_update_adapter(adapter);

		spin_lock(&adapter->adapter_lock);
		
		/* Continue scanning for another period. */
		mod_timer(&adapter->scan_timer, jiffies + SCAN_TIME);
		spin_unlock(&adapter->adapter_lock);
		
	} else {
		INFO("Scan on all channels has finished.\n");
		rt2x00_end_scan(adapter);
	}
}

void rt2x00_start_scan(struct _rt2x00_adapter *adapter, u8 flags)
{
	struct _scan_entry	*entry;
	struct _scan_entry	*next;
	u8			update = 0;

	write_lock(&adapter->config_lock);

	if(TEST_FLAG(adapter, ADAPTER_SCAN_IN_PROGRESS)){
		write_unlock(&adapter->config_lock);
		return;
	}

	if(TEST_FLAG(&adapter->scan, IW_SCAN_ALL_FREQ)){
		INFO("Start scanning all channels, switching to first channel...\n");

		/* We are going to scan all channels, so we start at the first channel. */
		adapter->config->channel = 0;

		/* We must update the adapter at some later point. */
		update = 1;
	} else {
		INFO("Start scanning current channel...\n");
	}

	write_unlock(&adapter->config_lock);

	/* Don't forget to load the new channel into the registers. */
	if(update)
		rt2x00_update_adapter(adapter);

	spin_lock(&adapter->adapter_lock);

	adapter->scan.flags = flags;	/* Update scan type. */

	/* Remove old scan list. */
	list_for_each_entry_safe(entry, next, &adapter->scan.result, list){
		list_del(&entry->list);
		kfree(entry);
	}


	INIT_LIST_HEAD(&adapter->scan.result);

	adapter->scan_timer.expires = jiffies + SCAN_TIME;
	adapter->scan_timer.data = (unsigned long) adapter;
	adapter->scan_timer.function = &rt2x00_scan_update;
	add_timer(&adapter->scan_timer);

	SET_FLAG(adapter, ADAPTER_SCAN_IN_PROGRESS);

	spin_unlock(&adapter->adapter_lock);
}

void rt2x00_end_scan(struct _rt2x00_adapter *adapter)
{

	if(!TEST_FLAG(adapter, ADAPTER_SCAN_IN_PROGRESS))
		return;

	spin_lock(&adapter->adapter_lock);

	adapter->scan.timestamp = jiffies;

	del_timer_sync(&adapter->scan_timer);

	CLEAR_FLAG(adapter, ADAPTER_SCAN_IN_PROGRESS);

	spin_unlock(&adapter->adapter_lock);

	INFO("Scan has ended.\n");
}

char rt2x00_duplicate_scan(struct _rt2x00_adapter *adapter, struct _scan_entry *scan_entry)
{
	struct _scan_entry	*entry;

	spin_lock(&adapter->adapter_lock);

	list_for_each_entry(entry, &adapter->scan.result, list){
		/* Check to see if we have found a duplicate. */
		if(memcmp(&scan_entry->bssid, &entry->bssid, sizeof(ETH_ALEN)) == 0)
			goto update;
	}

	spin_unlock(&adapter->adapter_lock);
	return 0;

update:
	/* We have found a duplicate scan entry. We onlyu have to update our old entry and we are done. */
	memcpy(scan_entry, entry, sizeof(struct _scan_entry));

	spin_unlock(&adapter->adapter_lock);

	return 1;
}

void rt2x00_update_scan_entry(struct _rt2x00_adapter *adapter, struct _scan_entry *scan_entry)
{
	/* 
	 * We add a simple wrapper around rt2x00_duplicate_scan(), 
	 * when calling rt2x00_update_scan_entry() we are not interested in the return value.
	 */
	if(rt2x00_duplicate_scan(adapter, scan_entry))
		return;
}

char rt2x00_validate_scan(struct _rt2x00_adapter *adapter, struct _scan_entry *scan_entry)
{
	char	status = 0;
	u8	counter = 0;

	read_lock(&adapter->config_lock);

	/* Scan only this essid. */
	if(!TEST_FLAG(&adapter->scan, IW_SCAN_ALL_ESSID))
		status += memcmp(scan_entry->essid, adapter->config->essid, IW_ESSID_MAX_SIZE);

	/* Scan only this channel. */
	if(!TEST_FLAG(&adapter->scan, IW_SCAN_ALL_FREQ))
		status += scan_entry->channel != adapter->config->channel;

	/* Scan only this mode. */
	if(!TEST_FLAG(&adapter->scan, IW_SCAN_ALL_MODE)){
		if(LINK_TYPE(adapter, IW_MODE_ADHOC))
			status = scan_entry->mode == IW_MODE_ADHOC;
		else
			status = scan_entry->mode == IW_MODE_MASTER;
	}

	if(!TEST_FLAG(&adapter->scan, IW_SCAN_ALL_RATE)){
		for(counter = 0; counter < MAX_BIT_RATES; counter++){
			/* If we have found our rate we are done with the check. */
			if(scan_entry->rates[counter] == BITRATE(adapter))
				goto exit;
		}
		status = 1; /* We haven't found the correct rate. */
	}

exit:
	read_unlock(&adapter->config_lock);
	return status;
}

void rt2x00_add_scan_entry(struct _rt2x00_adapter *adapter, struct _scan_entry *scan_entry)
{
	struct _scan_entry	*entry;
	struct _scan_entry	*new_entry = NULL;

	if(!TEST_FLAG(adapter, ADAPTER_SCAN_IN_PROGRESS)){
		ERROR("Adding a scan result while not busy scanning.\n");
		return;
	}

	if(rt2x00_validate_scan(adapter, scan_entry))
		return;

	if(rt2x00_duplicate_scan(adapter, scan_entry))
		return;

	DEBUG("Adding network to scan list\n");

	spin_lock(&adapter->adapter_lock);

	new_entry = kmalloc(sizeof(struct _scan_entry), GFP_ATOMIC);
	if(!new_entry){
		ERROR("Not enough memory to add scan entry.\n");
		spin_unlock(&adapter->adapter_lock);
		return;
	}

	memcpy(new_entry, scan_entry, sizeof(struct _scan_entry));

	list_for_each_entry(entry, &adapter->scan.result, list){
		/* We sort the list by rssi. */
		if(new_entry->rssi > entry->rssi){
			list_add(&new_entry->list, &entry->list);
			spin_unlock(&adapter->adapter_lock);
			return;
		}
	}

	/* We have the lowest rssi value of the list, so we add ourself at the end. */
	list_add_tail(&new_entry->list, &adapter->scan.result);

	spin_unlock(&adapter->adapter_lock);
}

static inline u8 rt2x00_check_scan_entry(struct _rt2x00_adapter *adapter, u8 search_mode, struct _scan_entry *scan_entry)
{
	if(search_mode == IW_MODE_AUTO || scan_entry->mode == search_mode){
		if(TEST_FLAG(adapter, FIX_BSSID))
			return memcmp(&adapter->config->bssid, &scan_entry->bssid, ETH_ALEN);
		else if(adapter->config->essid[0])
			return memcmp(&adapter->config->essid, &scan_entry->essid, IW_ESSID_MAX_SIZE);
		else
			return 0; /* Any essid, so always good. */
	}

	return -EINVAL;
}

/*
 * rt2x00_search_scan_entry must be called with the adapter_lock held.
 */
struct _scan_entry *rt2x00_search_scan_entry(struct _rt2x00_adapter *adapter, u8 search_mode)
{
	struct _scan_entry	*entry;

	if(TEST_FLAG(adapter, ADAPTER_SCAN_IN_PROGRESS)){
		NOTICE("Scan still in progress.\n");
		return NULL;
	}

	list_for_each_entry(entry, &adapter->scan.result, list){
		if(rt2x00_check_scan_entry(adapter, search_mode, entry))
			return entry;
	}

	WARNING("No valid scanning result found.\n");
	return NULL;
}

u32 rt2x00_create_mgmt_buffer(struct _rt2x00_adapter *adapter, void *buff, u16 ie_type)
{
	 void *start = buff;	/* Start address. */

	read_lock(&adapter->config_lock);

	if(ie_type & SET_IE_SSID){
		memset(buff, FRAME_IE_SSID, sizeof(u8));
		buff += sizeof(u8);
		memset(buff, strlen((u8*)adapter->config->essid), sizeof(u8));
		buff += sizeof(u8);
		memcpy(buff, adapter->config->essid, strlen((u8*)adapter->config->essid));
		buff += strlen((u8*)adapter->config->essid);
	}
	if(ie_type & SET_IE_RATES){
		memset(buff, FRAME_IE_RATES, sizeof(u8));
		buff += sizeof(u8);
		memcpy(buff, &adapter->capability->size_bit_rates, sizeof(u8));
		buff += sizeof(u8);
		memcpy(buff, adapter->capability->bit_rates, adapter->capability->size_bit_rates);
		buff += adapter->capability->size_bit_rates;
	}
	if(ie_type & SET_IE_FH){
		WARNING("Attempting to write unsupported IE into buffer (SET_IE_FH).\n");
		/*
		 * With unsupported I mean the following:
		 * I have no idea what information should be written into the buffer.
		 *
		 * memset(*buff, FRAME_IE_FH, sizeof(u8));
		 * buff += sizeof(u8);
		 */
	}
	if(ie_type & SET_IE_DS){
		memset(buff, FRAME_IE_DS, sizeof(u8));
		buff += sizeof(u8);
		memset(buff, sizeof(CHANNEL(adapter)), sizeof(u8));
		buff += sizeof(u8);
		memset(buff, CHANNEL(adapter), sizeof(u8));
		buff += sizeof(u8);
	}
	if(ie_type & SET_IE_CF){
		WARNING("Attempting to write unsupported IE into buffer (SET_IE_CF).\n");
		/*
		 * With unsupported I mean the following:
		 * I have no idea what information should be written into the buffer.
		 *
		 * memset(*buff, FRAME_IE_CF, sizeof(u8));
		 * buff += sizeof(u8);
		 */
	}
	if(ie_type & SET_IE_TIM){
		memset(buff, FRAME_IE_TIM, sizeof(u8));
		buff += sizeof(u8);
		memset(buff, sizeof(adapter->config->cfp_period), sizeof(u8));
		buff += sizeof(u8);
		memcpy(buff, &adapter->config->cfp_period, sizeof(u16));
		buff += sizeof(u16);
	}
	if(ie_type & SET_IE_IBSS){
		memset(buff, FRAME_IE_IBSS, sizeof(u8));
		buff += sizeof(u8);
		memset(buff, sizeof(adapter->config->atim_period), sizeof(u8));
		buff += sizeof(u8);
		memcpy(buff, &adapter->config->atim_period, sizeof(u16));
		buff += sizeof(u16);
	}

	read_unlock(&adapter->config_lock);

	return (buff - start);
}

void rt2x00_create_beacon(struct _rt2x00_adapter* adapter)
{
	struct _txd			txd;
	struct ieee802_11_mgmt_hdr	header_802_11;
	u32				length;

	spin_lock(&adapter->beacon_lock);
	read_lock(&adapter->config_lock);

	memset(&txd, 0x00, sizeof(struct _txd));
	memset(&header_802_11, 0x00, IEEE802_11_HLEN);

	header_802_11.frame_ctl = IEEE802_11_FTYPE_MGMT | IEEE802_11_STYPE_BEACON | IEEE802_11_FCTL_TODS;
	memset(&header_802_11.addr1, 0xff, ETH_ALEN);				/* Broadcast. */
	memcpy(&header_802_11.addr2, adapter->net_dev->dev_addr, ETH_ALEN);	/* Source address. */
	memcpy(&header_802_11.addr3, adapter->config->bssid, ETH_ALEN);		/* Destination address. */
	header_802_11.timestamp = jiffies;
	header_802_11.beacon_int = adapter->config->beacon;

	if(LINK_TYPE(adapter, IW_MODE_MASTER))
		header_802_11.capability |=  CAPABLE_ESS | CAPABLE_IBSS;
	else
		header_802_11.capability |=  CAPABLE_IBSS;

	if(TEST_FLAG(adapter, CRYPTO_ENABLED))
		header_802_11.capability |= CAPABLE_PRIVACY;
	if(adapter->config->preamble == PREAMBLE_SHORT)
		header_802_11.capability |= CAPABLE_SHORT_PREAMBLE;
	if(adapter->config->cfp_period)
		header_802_11.capability |= CAPABLE_CF_POLLABLE | CAPABLE_CF_POLL_REQ;
	if(adapter->config->slot_time == 9)
		header_802_11.capability |= CAPABLE_SHORT_SLOT_TIME;

	length = rt2x00_create_mgmt_buffer(adapter, (void*)&txd, SET_IE_SSID | SET_IE_RATES | SET_IE_DS | SET_IE_TIM | SET_IE_IBSS);

	txd.ifs = IFS_BACKOFF;

	cpu_to_device(adapter->dma->beacon_ring.data_addr, &header_802_11, sizeof(struct ieee802_11_mgmt_hdr));

	rt2x00_write_tx_descriptor(adapter, &txd, adapter->dma->beacon_ring.desc_addr, BASIC_RATE(adapter), length);

	read_unlock(&adapter->config_lock);
	spin_unlock(&adapter->beacon_lock);
}

void rt2x00_halt_beacon(struct _rt2x00_adapter* adapter)
{
	union _csr14	csr14;

	INFO("halt beacon generator.\n");

	rt2x00_register_read(adapter, CSR14, &csr14.word);

	csr14.field.beacon_gen = 0;

	rt2x00_register_write(adapter, CSR14, csr14.word);
}

void rt2x00_start_beacon(struct _rt2x00_adapter* adapter, u8 mode)
{
	union _csr12	csr12;
	union _csr14	csr14;
	union _bcncsr1	bcncsr1;

	rt2x00_halt_beacon(adapter); /* Halt current beacon, before starting a new beacon. */

	if(mode == IW_MODE_ADHOC || mode == IW_MODE_MASTER)
		rt2x00_create_beacon(adapter); /* Create beacon first, to prevent broadcast of incorrect beacons. */

	INFO("Start beacon generator.\n");

	rt2x00_register_read(adapter, CSR12, &csr12.word);
	csr12.field.beacon_interval = adapter->config->beacon / 16;	/* ASIC register in units of 1/16 TU */
	rt2x00_register_write(adapter, CSR12, csr12.word);


	rt2x00_register_read(adapter, BCNCSR1, &bcncsr1.word);
	if(mode ==  IW_MODE_INFRA)
		bcncsr1.field.preload = rt2x00_packet_duration(adapter, IEEE802_11_MGMT_HLEN, BASIC_RATE(adapter));
	else	/* we have delayed 1us clock to be 33*33.3Mhz, so preload this value to increase BEACON send ratio. */
		bcncsr1.field.preload = (adapter->config->beacon * 30) + 
					rt2x00_packet_duration(adapter, IEEE802_11_MGMT_HLEN, BASIC_RATE(adapter));
	rt2x00_register_write(adapter, BCNCSR1, bcncsr1.word);

	rt2x00_register_read(adapter, CSR14, &csr14.word);
	csr14.field.tsf_count = 1;
	csr14.field.tsf_sync = (mode ==  IW_MODE_ADHOC) ? 2 : 1;
	csr14.field.beacon_gen = ((mode ==  IW_MODE_ADHOC) || (mode == IW_MODE_MASTER)) ? 1 : 0;
	csr14.field.tbcn = 1;
	csr14.field.tbcn_preload = (adapter->config->beacon - 30) / 16;
	rt2x00_register_write(adapter, CSR14, csr14.word);
}

struct _scan_entry* rt2x00_receive_beacon(struct _rt2x00_adapter* adapter, struct sk_buff *skb)
{
	struct _scan_entry		*entry=NULL;
	struct ieee802_11_mgmt_hdr	*header_802_11 = (struct ieee802_11_mgmt_hdr*)skb->data;
	struct _beacon_ie		*ie = (struct _beacon_ie*) ((void*)(skb->data) + sizeof(struct ieee802_11_mgmt_hdr));
	struct _skb_cb_rx		*skb_cb = (struct _skb_cb_rx*)skb->cb;
	char 				*ie_data;
	
	if(!LINK_TYPE(adapter, IW_MODE_ADHOC) && !LINK_TYPE(adapter, IW_MODE_INFRA) && !LINK_TYPE(adapter, IW_MODE_AUTO))
		return NULL;

	entry=kmalloc(sizeof(struct _scan_entry),GFP_KERNEL);

	memset(entry, 0x00, sizeof(struct _scan_entry));

	memcpy(&entry->bssid, &header_802_11->addr2, ETH_ALEN);
	memcpy(&entry->timestamp, &header_802_11->timestamp, sizeof(unsigned long));
	memcpy(&entry->beacon, &header_802_11->beacon_int, sizeof(u64));
	memcpy(&entry->capability, &header_802_11->capability, sizeof(u16));

	if(adapter->chip_adapter == RT2400)
		entry->rssi = skb_cb->rxd.chip.rt2400.rssi;
	else
		entry->rssi = skb_cb->rxd.chip.rt2500.rssi;

	if(entry->capability & (CAPABLE_ESS | CAPABLE_IBSS))
		entry->mode = IW_MODE_MASTER;
	else
		entry->mode = IW_MODE_ADHOC;
	
	while((u8*)ie < skb->tail){
    		
		ie_data=&ie->data;
			
		switch(ie->frame){
			case FRAME_IE_SSID:
				if(ie->size > IW_ESSID_MAX_SIZE){
					WARNING("Essid length error in received beacon.\n");
				} else {
					memcpy(entry->essid, ie_data, ie->size);
				}
			break;
			case FRAME_IE_RATES:
				if(ie->size > MAX_BIT_RATES){
					WARNING("Rates length error in received beacon.\n");
				} else {
					memcpy(entry->rates, ie_data, ie->size);
				}
			break;
			case FRAME_IE_FH:
				WARNING("Unsupported IE (IE_FH) in received beacon.\n");
			break;
			case FRAME_IE_DS:
				if(ie->size != sizeof(u8) || *ie_data < 0 || *ie_data > adapter->capability->size_channels){
					WARNING("Illegal channel in received beacon.\n");
				} else {
					memcpy(&entry->channel, ie_data, ie->size);
				}
			break;
			case FRAME_IE_CF:
				if(ie->size < sizeof(u16)){
					WARNING("Illegal IE_CF length in received beacon.\n");
				} else {
					entry->cfp_count =  *ie_data;
					entry->cfp_period = *(ie_data + 1);
				}
			break;
			case FRAME_IE_TIM:
				if(ie->size < sizeof(u16)){
					WARNING("Illegal IE_TIM length in received beacon.\n");
				} else {
					entry->atim_count =  *(ie_data + 1);
					entry->atim_period = *(ie_data + 2);
				}
			break;
			case FRAME_IE_IBSS:
				if(ie->size != sizeof(u16)){
					WARNING("Illegal IE_IBSS length in received beacon.\n");
				} else {
					memcpy(&entry->atim_period, ie_data, ie->size);
				}
			break;
			default:
				/* Some AP's generate *loads* of other IE's in beacons this is
				not an error condition*/		
			break;
		}

		/*
		 * Increment pointer to next data segment.
		 * do (sizeof(_beacon_ie) - 1) to remove the _beacon_ie->data byte, which is added again with ie->size.
		 */

		ie = ((void*)ie)+ie->size+2; /* increase ptr by size of last data and two control bytes*/
	}


	return entry;
}

void rt2x00_send_ps_poll(struct _rt2x00_adapter *adapter)
{
	struct sk_buff		*skb = NULL;
	struct ieee802_11_hdr	*header802_11 = NULL;
	u32			size;

	/*
	 * Only send ps_poll when we are in Managed mode.
	 */
	if(!LINK_TYPE(adapter, IW_MODE_INFRA))
		return;

	/*
	 * Calculate the packet size.
	 * We only need the following data from the 802.11 header:
	 * frame_ctl, duration_id, addr1 and addr2.
	 */
	size = sizeof(u16) + sizeof(u16) + ETH_ALEN + ETH_ALEN;

	skb = dev_alloc_skb(size);
	if(!skb){
		ERROR("Not enough memory to send PS POLL!");
		return;
	}

	memset(skb->data, 0x00, size);

	header802_11 = (struct ieee802_11_hdr*)skb->data;

	header802_11->frame_ctl |= IEEE802_11_FTYPE_CTL | IEEE802_11_STYPE_PSPOLL | IEEE802_11_FCTL_TODS;

	header802_11->duration_id = adapter->config->sifs
				  + rt2x00_packet_duration(adapter, BASIC_RATE(adapter), size);

	memcpy(&header802_11->addr1, &adapter->config->bssid, ETH_ALEN);
	memcpy(&header802_11->addr2, &adapter->net_dev->dev_addr, ETH_ALEN);

	spin_lock_bh(&adapter->prio_lock);

	/* Enqueue packet for processing. */
	skb_queue_tail(&adapter->prio_packet_queue, skb);

	tasklet_schedule(&adapter->prio_tasklet);

	spin_unlock_bh(&adapter->prio_lock);
}

void rt2x00_send_null_frame(struct _rt2x00_adapter *adapter)
{
	struct sk_buff		*skb = NULL;
	struct ieee802_11_hdr	*header802_11 = NULL;
	u32			size;

	/*
	 * Only send null_frame when we are in Managed mode.
	 */
	if(!LINK_TYPE(adapter, IW_MODE_INFRA))
		return;

	/*
	 * Calculate the packet size.
	 * We need the entire 802.11 header, except the addr4.
	 */
	size = sizeof(struct ieee802_11_hdr) - ETH_ALEN;

	skb = dev_alloc_skb(size);
	if(!skb){
		ERROR("Not enough memory to send NULL FRAME!");
		return;
	}

	memset(skb->data, 0x00, size);

	header802_11 = (struct ieee802_11_hdr*)skb->data;

	header802_11->frame_ctl |= IEEE802_11_FTYPE_DATA | IEEE802_11_STYPE_NULLFUNC | IEEE802_11_FCTL_TODS;

	header802_11->duration_id = rt2x00_packet_duration(adapter, BASIC_RATE(adapter), size);

	memcpy(&header802_11->addr1, &adapter->config->bssid, ETH_ALEN);
	memcpy(&header802_11->addr2, &adapter->net_dev->dev_addr, ETH_ALEN);
	memcpy(&header802_11->addr3, &adapter->config->bssid, ETH_ALEN);

	spin_lock_bh(&adapter->prio_lock);

	/* Enqueue packet for processing. */
	skb_queue_tail(&adapter->prio_packet_queue, skb);

	tasklet_schedule(&adapter->prio_tasklet);

	spin_unlock_bh(&adapter->prio_lock);
}

char rt2x00_receive_mgmt_packet(struct _rt2x00_adapter *adapter, struct sk_buff *skb)
{
	struct ieee802_11_hdr		*header802_11 = (struct ieee802_11_hdr*)skb->data;
	struct _scan_entry		*scan_entry = (struct _scan_entry*)NULL;
	
	if(header802_11->frame_ctl & IEEE802_11_STYPE_BEACON){
		scan_entry = rt2x00_receive_beacon(adapter, skb);
		if(scan_entry!=NULL){
		    /* If we are not acively scanning we do wish to obtain the latest channel quality.*/
		    if(memcmp(scan_entry->bssid, adapter->config->bssid, ETH_ALEN) == 0)
			rt2x00_update_scan_entry(adapter, scan_entry);
			else if(TEST_FLAG(adapter, ADAPTER_SCAN_IN_PROGRESS))
				rt2x00_add_scan_entry(adapter, scan_entry);
		

	    /* Free the temp scan_entry buffer as we are done with it */
	    kfree(scan_entry);
	    }
	}

	return 0;
}

char rt2x00_transmit_mgmt(struct _rt2x00_adapter *adapter, struct sk_buff *skb)
{
	struct _txd		txd;
	u8			*dma = (u8*) adapter->dma->prio_ring[adapter->dma->prio_ring_index].data_addr;
	u8			*desc = (u8*) adapter->dma->prio_ring[adapter->dma->prio_ring_index].desc_addr;

	spin_lock_bh(&adapter->prio_lock);
	read_lock(&adapter->config_lock);

	if(rt2x00_test_free_descriptors(adapter, PRIO_RING, 1)){
		read_unlock(&adapter->config_lock);
		spin_unlock_bh(&adapter->tx_lock);
		return -ENOMEM;
	}

	device_to_cpu(&txd, adapter->dma->prio_ring[adapter->dma->prio_ring_index].desc_addr, sizeof(struct _txd));

	if(txd.owner == DESC_OWNED_BY_NIC || txd.valid){
		read_unlock(&adapter->config_lock);
		spin_unlock_bh(&adapter->tx_lock);
		return -ENOMEM;
	}

	cpu_to_device(dma, skb->data, skb->len);

	txd.more_frag = 0;
	txd.ifs = IFS_BACKOFF;
	txd.ack = 0;

	rt2x00_write_tx_descriptor(adapter, &txd, desc, BASIC_RATE(adapter), skb->len);

	INCREASE_RING(adapter->dma->prio_ring_index, PRIO_RING_SIZE);

	read_unlock(&adapter->config_lock);
	spin_unlock_bh(&adapter->prio_lock);

	return 0;
}
