/*
	rt2x00 SourceForge Project - <http://rt2x00.serialmonkey.com>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the
	Free Software Foundation, Inc.,
	59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

	Licensed under the GNU GPL
	Original code supplied under license from RaLink Inc, 2003
 */

/*
	Module Name: rt2x00_pci.c

	Abstract: PCI specific routines.

	Revision History:
	Who		When		What
	--------	-----------	-----------------------------
	IvD		<releasedate>	Initial Release.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/pci.h>

#include <net/iw_handler.h>

#include "rt2x00_adapter.h"

#ifndef CONFIG_PCI
#error Kernel has not been compiled with PCI support enabled!
#endif /* CONFIG_PCI */

void rt2x00_free_pci_dma(struct pci_dev *pci_dev, struct _device_mem *dma)
{
	/*
	 * Check every pointer to see if it is safe to free memory.
	 */
	if(dma->rx_ring.desc_addr)
		dma_free_coherent(pci_dev, RX_DESC_SIZE, dma->rx_ring.desc_addr, dma->rx_desc_dma);
	if(dma->rx_ring.data_addr)
		dma_free_coherent(pci_dev, RX_DATA_SIZE, dma->rx_ring.data_addr, dma->rx_data_dma);
	if(dma->tx_ring.desc_addr)
		dma_free_coherent(pci_dev, TX_DESC_SIZE, dma->tx_ring.desc_addr, dma->tx_desc_dma);
	if(dma->tx_ring.data_addr)
		dma_free_coherent(pci_dev, TX_DATA_SIZE, dma->tx_ring.data_addr, dma->tx_data_dma);
	if(dma->atim_ring.desc_addr)
		dma_free_coherent(pci_dev, ATIM_DESC_SIZE, dma->atim_ring.desc_addr, dma->atim_desc_dma);
	if(dma->atim_ring.data_addr)
		dma_free_coherent(pci_dev, ATIM_DATA_SIZE, dma->atim_ring.data_addr, dma->atim_data_dma);
	if(dma->prio_ring.desc_addr)
		dma_free_coherent(pci_dev, PRIO_DESC_SIZE, dma->prio_ring.desc_addr, dma->prio_desc_dma);
	if(dma->prio_ring.data_addr)
		dma_free_coherent(pci_dev, PRIO_DATA_SIZE, dma->prio_ring.data_addr, dma->prio_data_dma);
	if(dma->beacon_ring.desc_addr)
		dma_free_coherent(pci_dev, BEACON_DESC_SIZE, dma->beacon_ring.desc_addr, dma->beacon_desc_dma);
	if(dma->beacon_ring.data_addr)
		dma_free_coherent(pci_dev, BEACON_DATA_SIZE, dma->beacon_ring.data_addr, dma->beacon_data_dma);

	/*
	 * Release memory required for dma structure.
	 */
	kfree(dma);
}

struct _device_mem* rt2x00_init_pci_dma(struct pci_dev *pci_dev)
{
	struct _device_mem		*dma = NULL;

	dma = kmalloc(sizeof(struct _device_mem), GFP_ATOMIC);
	if(!dma){
		ERROR("Not enough memory to allocate DMA structure.\n");
		return NULL;
	}

	memset(dma, 0x00, sizeof(struct _device_mem));

	/*
	 * Allocate a DMA block per ring.
	 */
	dma->rx_ring.desc_addr = dma_alloc_coherent(pci_dev, RX_DESC_SIZE, &dma->rx_desc_dma, GFP_ATOMIC);
	dma->rx_ring.data_addr = dma_alloc_coherent(pci_dev, RX_DATA_SIZE, &dma->rx_data_dma, GFP_ATOMIC);
	dma->tx_ring.desc_addr = dma_alloc_coherent(pci_dev, TX_DESC_SIZE, &dma->tx_desc_dma, GFP_ATOMIC);
	dma->tx_ring.data_addr = dma_alloc_coherent(pci_dev, TX_DATA_SIZE, &dma->tx_data_dma, GFP_ATOMIC);
	dma->atim_ring.desc_addr = dma_alloc_coherent(pci_dev, ATIM_DESC_SIZE, &dma->atim_desc_dma, GFP_ATOMIC);
	dma->atim_ring.data_addr = dma_alloc_coherent(pci_dev, ATIM_DATA_SIZE, &dma->atim_data_dma, GFP_ATOMIC);
	dma->prio_ring.desc_addr = dma_alloc_coherent(pci_dev, PRIO_DESC_SIZE, &dma->prio_desc_dma, GFP_ATOMIC);
	dma->prio_ring.data_addr = dma_alloc_coherent(pci_dev, PRIO_DATA_SIZE, &dma->prio_data_dma, GFP_ATOMIC);
	dma->beacon_ring.desc_addr = dma_alloc_coherent(pci_dev, BEACON_DESC_SIZE, &dma->beacon_desc_dma, GFP_ATOMIC);
	dma->beacon_ring.data_addr = dma_alloc_coherent(pci_dev, BEACON_DATA_SIZE, &dma->beacon_data_dma, GFP_ATOMIC);

	/*
	 * Check if all allocations have succeeded.
	 */
	if(!dma->rx_ring.desc_addr || !dma->rx_ring.data_addr ||
	   !dma->tx_ring.desc_addr || !dma->tx_ring.data_addr ||
	   !dma->atim_ring.desc_addr || !dma->atim_ring.data_addr ||
	   !dma->prio_ring.desc_addr || !dma->prio_ring.data_addr ||
	   !dma->beacon_ring.desc_addr || !dma->beacon_ring.data_addr){
		/*
		 * We failed to acquire all dma rings.
		 * Free the succesfully allocated memory and exit.
		 */
		ERROR("Not enough memory to allocate DMA.\n");
		rt2x00_free_pci_dma(pci_dev, dma);
		return NULL;
	}

	/* Zero init ring descriptors */
	memset(dma->rx_ring.desc_addr, 0x00, RX_DESC_SIZE);
	memset(dma->tx_ring.desc_addr, 0x00, TX_DESC_SIZE);
	memset(dma->atim_ring.desc_addr, 0x00, ATIM_DESC_SIZE);
	memset(dma->prio_ring.desc_addr, 0x00, PRIO_DESC_SIZE);
	memset(dma->beacon_ring.desc_addr, 0x00, BEACON_DESC_SIZE);

	return dma;
}

static int rt2x00_pci_probe(struct pci_dev *pci_dev, const struct pci_device_id *id)
{
	struct net_device	*net_dev = NULL;
	struct _device_mem	*dma = NULL;
	void __iomem		*csr_addr;
	int			status = -ENODEV;

	if(id->driver_data != RT2400 && id->driver_data != RT2500){
		ERROR("Unknown device.\n");
		goto exit_fail;
	}

	if(pci_enable_device(pci_dev)){
		ERROR("Unable to enable device.\n");
		status = -EIO;
		goto exit_fail;
	}

	if(pci_set_dma_mask(pci_dev, DMA_64BIT_MASK)){
		if(pci_set_dma_mask(pci_dev, DMA_32BIT_MASK)){
			ERROR("PCI DMA not supported\n");
			goto exit_fail;
		}
	}

	if(pci_request_regions(pci_dev, pci_name(pci_dev))){
		ERROR("PCI request regions failed.\n");
		goto exit_fail;
	}

	csr_addr = ioremap(pci_resource_start(pci_dev, 0), pci_resource_len(pci_dev, 0));
	if (!csr_addr){
		ERROR("Ioremap failed.\n");
		goto exit_release_regions;
	}

	pci_set_master(pci_dev);

	/*
	 * Enables MWI - increases PCI performance if available.
	 */
	if(pci_set_mwi(pci_dev))
		ERROR("MWI not available\n");

	dma = rt2x00_init_pci_dma(pci_dev);
	if(!dma){
		ERROR("PCI DMA initialization failed.\n");
		status = -ENOMEM;
		goto exit_unmap;
	}

	net_dev = rt2x00_probe(id->driver_data, csr_addr, dma);
	if(!net_dev){
		ERROR("Net device allocation failure.\n");
		goto exit_free_dma;
	}

	SET_NETDEV_DEV(net_dev, &pci_dev->dev);
		
	status = register_netdev(net_dev);
	if(status){
		ERROR("register_netdev() failed\n");
		goto exit_free_dev; 
	}
	
	SET_MODULE_OWNER(net_dev);

	status = request_irq(pci_dev->irq, rt2x00_interrupt, SA_SHIRQ, net_dev->name, net_dev);
	if(status){
		ERROR("Allocation IRQ %d failed.\n", pci_dev->irq);
		goto exit_free_dev;
	}

	net_dev->irq = pci_dev->irq;

	pci_set_drvdata(pci_dev, net_dev);

	return 0;

exit_free_dev:
	rt2x00_remove(net_dev);

exit_free_dma:
	rt2x00_free_pci_dma(pci_dev, dma);

exit_unmap:
	pci_set_drvdata(pci_dev, NULL);
	iounmap(csr_addr);

exit_release_regions:
	pci_release_regions(pci_dev);

exit_fail:
	pci_disable_device(pci_dev);
	return status;
}

static void rt2x00_pci_remove(struct pci_dev *pci_dev)
{
	struct net_device	*net_dev = pci_get_drvdata(pci_dev);
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);

	rt2x00_remove(net_dev);

	rt2x00_free_pci_dma(pci_dev, adapter->dma_mem);

	free_irq(net_dev->irq, net_dev);
	pci_set_drvdata(pci_dev, NULL);

	if(adapter->register_base_address)
		iounmap(adapter->register_base_address);
	adapter->register_base_address = NULL;

	pci_release_regions(pci_dev);

	pci_disable_device(pci_dev);
}

/*
 * rt2x00_pci_suspend() & rt2x00_pci_resume()
 * The only pci specific code in the suspend and resume code is the location
 * of the adapter structure. So just pass that structure to the main suspend and resume code.
 */
static int rt2x00_pci_suspend(struct pci_dev *pci_dev, u32 state)
{
	struct net_device	*net_dev = pci_get_drvdata(pci_dev);
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);
	return rt2x00_suspend(adapter, state);
}

static int rt2x00_pci_resume(struct pci_dev *pci_dev)
{
	struct net_device	*net_dev = pci_get_drvdata(pci_dev);
	struct _rt2x00_adapter	*adapter = netdev_priv(net_dev);
	return rt2x00_resume(adapter);
}

/*
 * rt2x00 PCI device table, include all supported chipsets
 */
static struct pci_device_id rt2x00_pci_tbl[] = {
	{ PCI_DEVICE(0x1814, 0x0101), .driver_data = RT2400},
	{ PCI_DEVICE(0x1814, 0x0201), .driver_data = RT2500},
	{0,}
};

static char version[] = DRV_NAME_PCI " - " DRV_VERSION " (" DRV_RELDATE ") by " DRV_PROJECT "\n";

MODULE_AUTHOR(DRV_PROJECT);
MODULE_VERSION(DRV_VERSION);
MODULE_DESCRIPTION("Ralink RT2400/RT2500 PCI & PCMCIA Wireless LAN driver. Released: " DRV_RELDATE);
MODULE_SUPPORTED_DEVICE("Ralink RT2400 & RT2500 PCI & PCMCIA chipset based cards");
MODULE_DEVICE_TABLE(pci, rt2x00_pci_tbl);
MODULE_LICENSE("GPL");

/*
 * Our PCI driver structure
 */
static struct pci_driver rt2x00_pci_driver = {
	.name		= DRV_NAME_PCI,
	.id_table	= rt2x00_pci_tbl,
	.probe		= rt2x00_pci_probe,
	.remove		= __devexit_p(rt2x00_pci_remove),
	.suspend	= rt2x00_pci_suspend,
	.resume		= rt2x00_pci_resume,
};

static int rt2x00_pci_init(void)
{
	printk(KERN_INFO "Loading module: %s\n", version);
	return pci_register_driver(&rt2x00_pci_driver);
}

static void rt2x00_pci_exit(void)
{
	printk(KERN_INFO "Unloading module: %s\n", version);
	pci_unregister_driver(&rt2x00_pci_driver);
}

module_init(rt2x00_pci_init);
module_exit(rt2x00_pci_exit);
