/*
	rt2x00 SourceForge Project - <http://rt2x00.serialmonkey.com>

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the
	Free Software Foundation, Inc.,
	59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

	Licensed under the GNU GPL
	Original code supplied under license from RaLink Inc, 2003
 */

/*
	Module Name: rt2x00_usb.c

	Abstract: USB specific routines.

	Revision History:
	Who		When		What
	--------	-----------	-----------------------------
	IvD		<releasedate>	Initial Release.
 */

#include <linux/kernel.h>
#include <linux/module.h>
#include <linux/interrupt.h>
#include <linux/spinlock.h>
#include <linux/list.h>
#include <linux/skbuff.h>
#include <linux/netdevice.h>
#include <linux/etherdevice.h>
#include <linux/usb.h>

#include <net/iw_handler.h>

#include "rt2x00_adapter.h"

#if !defined(CONFIG_USB) && !defined(CONFIG_USB_MODULE)
#error Kernel has not been compiled with USB support enabled!
#endif /* CONFIG_USB */

static void* rt2x00_usb_probe(struct usb_device *usb_dev, unsigned intf, const struct usb_device_id *id)
{
	if(id->driver_info != RT2500){
		ERROR("Unknown device.\n");
		goto exit_fail;
	}

	usb_get_dev(usb_dev);








exit_fail:
	usb_put_dev(usb_dev);

	return NULL;
}

static void rt2x00_usb_disconnect(struct usb_device *usb_dev, void *dev)
{
	struct net_device	*net_dev = (struct net_device*) dev;

	unregister_netdev(net_dev);

	rt2x00_remove(netdev_priv(net_dev));

	free_irq(net_dev->irq, net_dev);

	iounmap((void *)(net_dev->base_addr));
	
	free_netdev(net_dev);

	usb_put_dev(usb_dev);
}

/* 
 * Following part is just for 2.6.x kernels, since the USB API has changed between 2.4 and 2.6,
 * we have to use the following wrappers to convert the 2.6.x API to the 2.4.x API.
 */
#if LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0)
static int rt2x00_usb_probe_26(struct usb_interface *usb_intf, const struct usb_device_id *id)
{
	if(!rt2x00_usb_probe(interface_to_usbdev(usb_intf), 0,id))
		return -ENODEV;
	return 0;
}

static void rt2x00_usb_disconnect_26(struct usb_interface *usb_intf)
{
	rt2x00_usb_disconnect(interface_to_usbdev(usb_intf), usb_get_intfdata(usb_intf));
}

/*
 * rt2x00_usb_suspend() & rt2x00_usb_resume()
 * The only usb specific code in the suspend and resume code is the location
 * of the adapter structure. So just pass that structure to the main suspend and resume code.
 */
static int rt2x00_usb_suspend(struct usb_interface *usb_dev, u32 state)
{
	return rt2x00_suspend(usb_get_intfdata(usb_dev), state);
}

static int rt2x00_usb_resume(struct usb_interface *usb_dev)
{
	return rt2x00_resume(usb_get_intfdata(usb_dev));
}
#endif /* LINUX_VERSION_CODE > KERNEL_VERSION(2,6,0) */

/*
 * rt2x00 USB device table, include all supported chipsets
 */
static struct usb_device_id rt2x00_usb_tbl[] = {
	{ USB_DEVICE(0x1814, 0x0201), .driver_info = RT2500},
	{0,}
};

static char version[] = DRV_NAME_USB " - " DRV_VERSION " (" DRV_RELDATE ") by " DRV_PROJECT "\n";

MODULE_AUTHOR(DRV_PROJECT);
MODULE_VERSION(DRV_VERSION);
MODULE_DESCRIPTION("Ralink RT2500 USB Wireless LAN driver. Released: " DRV_RELDATE);
MODULE_SUPPORTED_DEVICE("Ralink RT2500 USB chipset based cards");
MODULE_DEVICE_TABLE(usb, rt2x00_usb_tbl);
MODULE_LICENSE("GPL");

/*
 * Our USB driver structure
 */

static struct usb_driver rt2x00_usb_driver = {
	.owner		= THIS_MODULE,
	.name		= DRV_NAME_USB,
	.id_table	= rt2x00_usb_tbl,
#if LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0)
	.probe		= rt2x00_usb_probe,
	.disconnect	= rt2x00_usb_disconnect,
#else /* LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) */
	.probe		= rt2x00_usb_probe_26,
	.disconnect	= rt2x00_usb_disconnect_26,
	.suspend	= rt2x00_usb_suspend,
	.resume		= rt2x00_usb_resume,
#endif /* LINUX_VERSION_CODE < KERNEL_VERSION(2,6,0) */
};

static int rt2x00_usb_init(void)
{
	printk(KERN_INFO "Loading module: %s", version);
	return usb_register(&rt2x00_usb_driver);
}

static void rt2x00_usb_exit(void)
{
	printk(KERN_INFO "Unloading module: %s", version);
	usb_deregister(&rt2x00_usb_driver);
}

module_init(rt2x00_usb_init);
module_exit(rt2x00_usb_exit);
